# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import json
from datetime import datetime
from pathlib import Path
from typing import Any

import aiosqlite

from gallia.command.config import GalliaBaseModel
from gallia.db.log import LogMode
from gallia.log import get_logger
from gallia.services.uds.core.service import (
    PositiveResponse,
    SubFunctionRequest,
    SubFunctionResponse,
    UDSRequest,
    UDSResponse,
)
from gallia.services.uds.core.utils import bytes_repr as bytes_repr_
from gallia.services.uds.ecu import ECUProperties
from gallia.utils import handle_task_error, set_task_handler_ctx_variable


def bytes_repr(data: bytes | bytearray) -> str:
    return bytes_repr_(data, False, None)


schema_version = "4.0"

DB_SCHEMA = f"""
CREATE TABLE IF NOT EXISTS version (
  schema text unique,
  version text
);
CREATE TABLE IF NOT EXISTS ecu (
  id integer primary key,
  name text,
  oem text,
  manufacturer text,
  oem_uid text
);
CREATE TABLE IF NOT EXISTS address (
  id integer primary key,
  url text not null unique,
  ecu int references ecu(id) on update cascade on delete set null
);
CREATE TABLE IF NOT EXISTS run_meta (
  id integer primary key,
  script text not null,
  config json not null check(json_valid(config)),
  start_time real not null,
  start_timezone text not null,
  end_time real,
  end_timezone text check((end_timezone is null) = (end_time is null)),
  exit_code int,
  path text,
  exclude boolean
);
CREATE TABLE IF NOT EXISTS error_log (
  level int not null,
  time real not null,
  timezone text not null,
  message text not null,
  meta int references run_meta(id) not null
);
CREATE TABLE IF NOT EXISTS discovery_run (
  id integer primary key,
  protocol text not null,
  meta int references run_meta(id) on update cascade on delete cascade
);
CREATE TABLE IF NOT EXISTS scan_run (
  id integer primary key,
  address int references address(id) on update cascade on delete set null,
  properties_pre json check(properties_pre is null or json_valid(properties_pre)),
  properties_post json check(properties_post is null or json_valid(properties_post)),
  meta int references run_meta(id) on update cascade on delete cascade
);
CREATE TABLE IF NOT EXISTS discovery_result (
  id integer primary key,
  run int not null references discovery_run(id) on update cascade on delete cascade,
  address int not null references address(id) on update cascade on delete cascade
);
CREATE TABLE IF NOT EXISTS scan_result (
  id integer primary key,
  run int not null references scan_run(id) on update cascade on delete cascade,
  log_mode text not null check(log_mode in ('implicit', 'explicit', 'emphasized')),
  state json check(state is null or json_valid(state)),
  request_pdu blob not null,
  request_time real not null,
  request_timezone text not null,
  request_data json check(json_valid(request_data)) not null,
  response_pdu blob,
  response_time real,
  response_timezone text check((response_timezone is null) = (response_time is null)),
  response_data json check(response_data is null or json_valid(response_data)),
  exception text
);
CREATE INDEX IF NOT EXISTS ix_scan_result_request_pdu ON scan_result(request_pdu);
CREATE TABLE IF NOT EXISTS session_transition (
  run int not null references scan_run(id) on update cascade on delete cascade,
  destination int not null,
  steps json check(json_valid(steps))
);

CREATE VIEW IF NOT EXISTS run_stats AS
SELECT
  ru.id AS run,
  ecu.name AS ECU,
  rm.script,
  rm.arguments,
  strftime('%Y-%m-%d %H:%M:%f', rm.start_time, 'unixepoch', 'localtime') AS start,
  strftime('%Y-%m-%d %H:%M:%f', rm.end_time, 'unixepoch', 'localtime') AS end,
  cast((CASE WHEN end_time IS NULL THEN strftime('%s','now') ELSE end_time END - start_time) / 86400 AS int) || ' ' ||
    time(CASE WHEN end_time IS NULL THEN strftime('%s','now') ELSE end_time END - start_time, 'unixepoch') AS duration,
  exit_code,
  ru.properties_pre = ru.properties_post AS equal_props,
  n_msgs
FROM
  run_meta rm,
  scan_run ru LEFT JOIN (SELECT run, count(*) AS n_msgs FROM scan_result GROUP BY run) sc ON ru.id = sc.run,
  address ad LEFT JOIN ecu ON ecu.id = ad.ecu
WHERE ru.meta = rm.id
AND ad.id = ru.address
GROUP BY ru.id;


INSERT OR IGNORE INTO version VALUES('main', '{schema_version}');
"""

logger = get_logger(__name__)


class DBHandler:
    def __init__(self, database: Path):
        self.tasks: list[asyncio.Task[None]] = []
        self.path = database
        self.connection: aiosqlite.Connection | None = None
        self.scan_run: int | None = None
        self.target: str | None = None
        self.discovery_run: int | None = None
        self.meta: int | None = None
        self._executor_task: asyncio.Task[None] | None = None
        self._execute_queue: asyncio.Queue[tuple[str, tuple[Any, ...]]] | None = None

    async def connect(self) -> None:
        """It is important that `connect` and `disconnect` are called in the same event loop!"""

        if self.connection is not None:
            logger.warning("Already connected to the database")
            return

        self.path.parent.mkdir(exist_ok=True, parents=True)
        self.connection = await aiosqlite.connect(self.path)
        await self.connection.execute("PRAGMA foreign_keys = 1")

        # Allows to read the database in parallel to a scan without causing delays or even losing data
        # This setting is persistent for the database and leads to the creation of extra files
        # See https://www.sqlite.org/wal.html for further information
        await self.connection.execute("PRAGMA journal_mode = WAL")

        await self.connection.execute("PRAGMA busy_timeout = 10000")

        await self.connection.executescript(DB_SCHEMA)
        await self.check_version()

        # This queue is meant to be used for usage-heavy executes that are not time-sensitive, e.g. UDS messages
        self._execute_queue = asyncio.Queue()
        self._executor_task = asyncio.create_task(self._executor_func())
        self._executor_task.add_done_callback(
            handle_task_error,
            context=set_task_handler_ctx_variable(__name__, "DbHandler"),
        )

    async def _executor_func(self) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self._execute_queue is not None, "Queue was not started"

        try:
            while True:
                (query, query_parameter) = await self._execute_queue.get()

                try:
                    await self.connection.execute(query, query_parameter)
                    await self.connection.commit()
                except aiosqlite.OperationalError:
                    logger.warning(
                        f"Could not log message for {query_parameter[5]} to database. Retrying ..."
                    )
                    # TODO: This could lead to an infinite loop when there are recurring OperationalErrors!
                    await self._execute_queue.put((query, query_parameter))
                finally:
                    # Inform the the queue that the query was fully processed to track progress
                    self._execute_queue.task_done()

        except asyncio.CancelledError:
            logger.debug("Database worker cancelled")
        except asyncio.IncompleteReadError as e:
            logger.debug(f"Database worker received EOF: {e}")
        except Exception as e:
            logger.critical(f"Database worker died: {e!r}")

    async def disconnect(self) -> None:
        """It is important that `connect` and `disconnect` are called in the same event loop!"""

        assert self.connection is not None, "Not connected to the database"
        assert self._execute_queue is not None, "Queue is already detached"
        assert self._executor_task is not None, "Task is already detached"

        logger.info("Syncing database…")
        try:
            # Wait for all queries in the queue to be written to the database and cancel task afterwards.
            # TODO: this could block infinitely if there are OperationalErrors writing to the database in
            # the `_executor_func()`
            await self._execute_queue.join()
            self._executor_task.cancel()
            await self._executor_task
        except Exception as e:
            logger.error(f"Could not properly clean up the database task: {e!r}")
        finally:
            self._execute_queue = None
            self._executor_task = None

        try:
            await self.connection.commit()
        finally:
            await self.connection.close()
            self.connection = None
        logger.info("Database closed")

    async def check_version(self) -> None:
        assert self.connection is not None, "Not connected to the database"

        query = 'SELECT version FROM version WHERE schema = "main"'
        cursor: aiosqlite.Cursor = await self.connection.execute(query)
        row = await cursor.fetchone()

        if row is None:
            raise ValueError("No schema version is specified in the database!")

        version = row[0]

        # This could be a much more complex version to support some older version as well
        # Additionally one might bind certain features to this version or trigger a database migration, etc
        if version != schema_version:
            raise ValueError(
                f"The version of the database schema is not supported! ({version} != {schema_version})"
            )

    async def insert_run_meta(  # noqa: PLR0913
        self,
        script: str,
        config: GalliaBaseModel,
        start_time: datetime,
        path: Path | None,
    ) -> None:
        assert self.connection is not None, "Not connected to the database"

        query = (
            "INSERT INTO "
            "run_meta(script, config, start_time, start_timezone, path, exclude) "
            "VALUES (?, ?, ?, ?, ?, FALSE)"
        )
        cursor = await self.connection.execute(
            query,
            (
                script,
                config.model_dump_json(),
                start_time.timestamp(),
                start_time.tzname(),
                str(path),
            ),
        )

        self.meta = cursor.lastrowid

        await self.connection.commit()

    async def complete_run_meta(
        self, end_time: datetime, exit_code: int, path: Path | None
    ) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.meta is not None, "Run meta not yet created"

        query = "UPDATE run_meta SET end_time = ?, end_timezone = ?, exit_code = ?, path = ? WHERE id = ?"
        await self.connection.execute(
            query,
            (end_time.timestamp(), end_time.tzname(), exit_code, str(path), self.meta),
        )
        await self.connection.commit()

    async def insert_scan_run(self, target: str) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.meta is not None, "Run meta not yet created"

        await self.connection.execute("INSERT OR IGNORE INTO address(url) VALUES(?)", (target,))

        query = (
            "INSERT INTO scan_run(address, meta) VALUES ((SELECT id FROM address WHERE url = ?), ?)"
        )
        cursor = await self.connection.execute(query, (target, self.meta))

        self.scan_run = cursor.lastrowid
        self.target = target
        await self.connection.commit()

    async def insert_scan_run_properties_pre(self, properties_pre: ECUProperties) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.scan_run is not None, "Scan run not yet created"

        query = "UPDATE scan_run SET properties_pre = ? WHERE id = ?"
        await self.connection.execute(query, (properties_pre.to_json(), self.scan_run))
        await self.connection.commit()

    async def complete_scan_run(self, properties_post: ECUProperties) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.scan_run is not None, "Scan run not yet created"

        query = "UPDATE scan_run SET properties_post = ? WHERE id = ?"
        await self.connection.execute(query, (properties_post.to_json(), self.scan_run))
        await self.connection.commit()

    async def insert_discovery_run(self, protocol: str) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.meta is not None, "Run meta not yet created"

        query = "INSERT INTO discovery_run(protocol, meta) VALUES (?, ?)"
        cursor = await self.connection.execute(query, (protocol, self.meta))
        self.discovery_run = cursor.lastrowid
        await self.connection.commit()

    async def insert_discovery_result(self, target: str) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self.discovery_run is not None, "Discovery run not yet created"

        await self.connection.execute("INSERT OR IGNORE INTO address(url) VALUES(?)", (target,))

        query = "INSERT INTO discovery_result(address, run) VALUES ((SELECT id FROM address WHERE url = ?), ?)"

        await self.connection.execute(query, (target, self.discovery_run))
        await self.connection.commit()

    async def insert_scan_result(  # noqa: PLR0913
        self,
        state: dict[str, Any],
        request: UDSRequest,
        response: UDSResponse | None,
        exception: Exception | None,
        send_time: datetime,
        receive_time: datetime | None,
        log_mode: LogMode,
    ) -> None:
        assert self.connection is not None, "Not connected to the database"
        assert self._execute_queue is not None, "Queue not yet created"
        assert self.scan_run is not None, "Scan run not yet created"

        request_attributes: dict[str, Any] = {
            "service_id": request.service_id,
            "data": bytes_repr(request.data),
        }
        response_attributes: dict[str, Any] = {}

        if isinstance(request, SubFunctionRequest):
            request_attributes["sub_function"] = request.sub_function

        for attr, value in request.__dict__.items():
            if not attr.startswith("_"):
                request_attributes[attr] = value

                if isinstance(value, bytes | bytearray):
                    request_attributes[attr] = bytes_repr(value)
                elif (
                    isinstance(value, list)
                    and len(value) > 0
                    and isinstance(value[0], bytes | bytearray)
                ):
                    request_attributes[attr] = [bytes_repr(v) for v in value]

        if response is not None:
            response_attributes = {"service_id": response.service_id}

            if isinstance(response, PositiveResponse):
                response_attributes["data"] = bytes_repr(response.data)

            if isinstance(response, SubFunctionResponse):
                response_attributes["sub_function"] = response.sub_function

            for attr, value in response.__dict__.items():
                if not attr.startswith("_") and attr not in ["trigger_request"]:
                    response_attributes[attr] = value

                    if isinstance(value, bytes | bytearray):
                        response_attributes[attr] = bytes_repr(value)
                    elif (
                        isinstance(value, list)
                        and len(value) > 0
                        and isinstance(value[0], bytes | bytearray)
                    ):
                        response_attributes[attr] = [bytes_repr(v) for v in value]

        query = (
            "INSERT INTO scan_result(run, state, request_pdu, request_time, request_timezone, request_data, "
            "response_pdu, response_time, response_timezone, response_data, exception, log_mode) "
            "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"
        )

        # This has do be done here, in order to make sure, that only "immutable" objects are passed to a different task
        query_parameter = (
            self.scan_run,
            json.dumps(state),
            bytes_repr(request.pdu),
            send_time.timestamp(),
            send_time.tzname(),
            json.dumps(request_attributes),
            bytes_repr(response.pdu) if response is not None else None,
            receive_time.timestamp() if response is not None and receive_time is not None else None,
            receive_time.tzname() if response is not None and receive_time is not None else None,
            json.dumps(response_attributes) if response is not None else None,
            repr(exception) if exception is not None else None,
            log_mode.name,
        )

        await self._execute_queue.put((query, query_parameter))

    async def insert_session_transition(self, destination: int, steps: list[int]) -> None:
        assert self.connection is not None, "Not connected to the database"

        query = "INSERT INTO session_transition VALUES(?, ?, ?)"
        parameters = (self.scan_run, destination, json.dumps(steps))
        await self.connection.execute(query, parameters)

    async def get_sessions(self) -> list[int]:
        assert self.connection is not None, "Not connected to the database"
        assert self.target is not None, "Scan run not yet created, target unknown"

        query = (
            "SELECT DISTINCT destination "
            "FROM session_transition st, "
            "     scan_run sr, "
            "     address ad "
            "WHERE st.run = sr.id AND sr.address = ad.id "
            "AND ad.url = ?"
        )
        parameters = (self.target,)

        cursor: aiosqlite.Cursor = await self.connection.execute(query, parameters)
        return [x[0] for x in await cursor.fetchall()]

    async def get_session_transition(self, destination: int) -> list[int] | None:
        assert self.connection is not None, "Not connected to the database"
        assert self.target is not None, "Scan run not yet created, target unknown"

        query = (
            "SELECT steps "
            "FROM session_transition st, "
            "     scan_run sr, "
            "     address ad "
            "WHERE st.run = sr.id AND sr.address = ad.id "
            "AND st.destination = ? AND ad.url = ?"
        )
        parameters = (destination, self.target)
        cursor: aiosqlite.Cursor = await self.connection.execute(query, parameters)
        row = await cursor.fetchone()

        if row is None:
            return None

        result: list[int] = json.loads(row[0])
        return result



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import socket as s
import struct
import sys
import time
from dataclasses import dataclass
from typing import Self

assert sys.platform.startswith("linux"), "unsupported platform"

from pydantic import BaseModel, field_validator

from gallia.log import get_logger
from gallia.transports._can_constants import (
    CAN_EFF_FLAG,
    CAN_EFF_MASK,
    CAN_ERR_FLAG,
    CAN_HEADER_FMT,
    CAN_INV_FILTER,
    CAN_RAW,
    CAN_RAW_FD_FRAMES,
    CAN_RAW_FILTER,
    CAN_RAW_JOIN_FILTERS,
    CAN_RTR_FLAG,
    CAN_SFF_MASK,
    CANFD_BRS,
    CANFD_ESI,
    CANFD_MTU,
    SOL_CAN_RAW,
)
from gallia.transports.base import BaseTransport, TargetURI
from gallia.utils import auto_int

logger = get_logger(__name__)


@dataclass(kw_only=True, slots=True, frozen=True)
class CANMessage:
    timestamp: float = 0.0
    arbitration_id: int = 0

    # TODO: Add a frametype attribute?
    is_extended_id: bool = False
    is_remote_frame: bool = False
    is_error_frame: bool = False

    dlc: int | None = None
    data: bytes = b""
    is_fd: bool = False
    is_rx: bool = True
    bitrate_switch: bool = False
    error_state_indicator: bool = False

    # TODO: Is this semantically correct? Check this.
    def __len__(self) -> int:
        if self.dlc is None:
            return len(self.data)
        return self.dlc

    def _compose_arbitration_id(self) -> int:
        can_id = self.arbitration_id
        if self.is_extended_id:
            can_id |= CAN_EFF_FLAG
        if self.is_remote_frame:
            can_id |= CAN_RTR_FLAG
        if self.is_error_frame:
            can_id |= CAN_ERR_FLAG
        return can_id

    def pack(self) -> bytes:
        can_id = self._compose_arbitration_id()
        flags = 0
        if self.bitrate_switch:
            flags |= CANFD_BRS
        if self.error_state_indicator:
            flags |= CANFD_ESI
        max_len = 64 if self.is_fd else 8
        data = bytes(self.data).ljust(max_len, b"\x00")
        return CAN_HEADER_FMT.pack(can_id, len(self), flags) + data

    @staticmethod
    def _dissect_can_frame(frame: bytes) -> tuple[int, int, int, bytes]:
        can_id, can_dlc, flags = CAN_HEADER_FMT.unpack_from(frame)
        if len(frame) != CANFD_MTU:
            # Flags not valid in non-FD frames
            flags = 0
        return can_id, can_dlc, flags, frame[8 : 8 + can_dlc]

    @classmethod
    def unpack(cls, frame: bytes) -> Self:
        can_id, can_dlc, flags, data = cls._dissect_can_frame(frame)

        # EXT, RTR, ERR flags -> boolean attributes
        #   /* special address description flags for the CAN_ID */
        #   #define CAN_EFF_FLAG 0x80000000U /* EFF/SFF is set in the MSB */
        #   #define CAN_RTR_FLAG 0x40000000U /* remote transmission request */
        #   #define CAN_ERR_FLAG 0x20000000U /* error frame */
        is_extended_frame_format = bool(can_id & CAN_EFF_FLAG)
        is_remote_transmission_request = bool(can_id & CAN_RTR_FLAG)
        is_error_frame = bool(can_id & CAN_ERR_FLAG)
        is_fd = len(frame) == CANFD_MTU
        bitrate_switch = bool(flags & CANFD_BRS)
        error_state_indicator = bool(flags & CANFD_ESI)

        if is_extended_frame_format:
            arbitration_id = can_id & CAN_EFF_MASK
        else:
            arbitration_id = can_id & CAN_SFF_MASK

        return cls(
            arbitration_id=arbitration_id,
            is_extended_id=is_extended_frame_format,
            is_remote_frame=is_remote_transmission_request,
            is_error_frame=is_error_frame,
            is_fd=is_fd,
            bitrate_switch=bitrate_switch,
            error_state_indicator=error_state_indicator,
            dlc=can_dlc,
            data=data,
        )


class RawCANConfig(BaseModel):
    is_extended: bool = False
    is_fd: bool = False
    dst_id: int | None = None

    @field_validator(
        "dst_id",
        mode="before",
    )
    def auto_int(cls, v: str) -> int:
        return auto_int(v)


class RawCANTransport(BaseTransport, scheme="can-raw"):
    def __init__(self, target: TargetURI, config: RawCANConfig, sock: s.socket) -> None:
        super().__init__(target)

        self._sock = sock
        self.config = config

    @classmethod
    async def connect(
        cls,
        target: str | TargetURI,
        timeout: float | None = None,
    ) -> Self:
        t = target if isinstance(target, TargetURI) else TargetURI(target)
        cls.check_scheme(t)

        if t.hostname is None:
            raise ValueError("empty interface")

        sock = s.socket(s.PF_CAN, s.SOCK_RAW, CAN_RAW)
        sock.bind((t.hostname,))
        config = RawCANConfig.model_validate(t.qs_flat)

        if config.is_fd is True:
            sock.setsockopt(SOL_CAN_RAW, CAN_RAW_FD_FRAMES, 1)

        sock.setblocking(False)

        return cls(t, config, sock)

    def set_filter(self, can_ids: list[int], inv_filter: bool = False) -> None:
        if not can_ids:
            return
        filter_mask = CAN_EFF_MASK if self.config.is_extended else CAN_SFF_MASK
        data = b""
        for can_id in can_ids:
            if inv_filter:
                can_id |= CAN_INV_FILTER  # noqa: PLW2901
            data += struct.pack("@II", can_id, filter_mask)
        self._sock.setsockopt(SOL_CAN_RAW, CAN_RAW_FILTER, data)
        if inv_filter:
            self._sock.setsockopt(SOL_CAN_RAW, CAN_RAW_JOIN_FILTERS, 1)

    async def read(
        self,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> bytes:
        raise RuntimeError("RawCANTransport is a special snowflake")

    async def write(
        self,
        data: bytes,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> int:
        if self.config.dst_id:
            return await self.sendto(data, self.config.dst_id, timeout, tags)
        raise ValueError("dst_id not set")

    async def sendto(
        self,
        data: bytes,
        dst: int,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> int:
        msg = CANMessage(
            arbitration_id=dst,
            data=data,
            is_extended_id=self.config.is_extended,
            is_fd=self.config.is_fd,
        )
        t = tags + ["write"] if tags is not None else ["write"]
        if self.config.is_extended:
            logger.trace(f"{dst:08x}#{data.hex()}", extra={"tags": t})
        else:
            logger.trace(f"{dst:03x}#{data.hex()}", extra={"tags": t})

        loop = asyncio.get_running_loop()
        await asyncio.wait_for(loop.sock_sendall(self._sock, msg.pack()), timeout)
        return len(data)

    async def recvfrom(
        self,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> tuple[int, bytes]:
        loop = asyncio.get_running_loop()
        can_frame = await asyncio.wait_for(loop.sock_recv(self._sock, self.BUFSIZE), timeout)
        msg = CANMessage.unpack(can_frame)

        t = tags + ["read"] if tags is not None else ["read"]
        if msg.is_extended_id:
            logger.trace(f"{msg.arbitration_id:08x}#{msg.data.hex()}", extra={"tags": t})
        else:
            logger.trace(f"{msg.arbitration_id:03x}#{msg.data.hex()}", extra={"tags": t})
        return msg.arbitration_id, msg.data

    async def close(self) -> None:
        pass

    async def get_idle_traffic(self, sniff_time: float) -> list[int]:
        """Listen to traffic on the bus and return list of IDs
        which are seen in the specified period of time.
        The output of this function can be used as input to set_filter.
        """
        addr_idle: list[int] = []
        t1 = time.time()
        while time.time() - t1 < sniff_time:
            try:
                addr, _ = await self.recvfrom(timeout=1)
                if addr not in addr_idle:
                    logger.info(f"Received a message from {addr:03x}")
                    addr_idle.append(addr)
            except TimeoutError:
                continue
        addr_idle.sort()
        return addr_idle



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import sys
from typing import Self

from pydantic import model_validator

assert sys.platform.startswith("linux"), "unsupported platform"

from gallia.command import UDSDiscoveryScanner
from gallia.command.config import AutoInt, Field, HexBytes
from gallia.command.uds import UDSDiscoveryScannerConfig
from gallia.log import get_logger
from gallia.services.uds import NegativeResponse, UDSClient, UDSRequest
from gallia.services.uds.core.utils import g_repr
from gallia.transports import ISOTPTransport, RawCANTransport, TargetURI
from gallia.utils import can_id_repr

logger = get_logger(__name__)


class IsotpDiscovererConfig(UDSDiscoveryScannerConfig):
    start: AutoInt = Field(description="set start address", metavar="INT")
    stop: AutoInt = Field(description="set end address", metavar="INT")
    padding: AutoInt | None = Field(None, description="set isotp padding")
    pdu: HexBytes = Field(bytes([0x3E, 0x00]), description="set pdu used for discovery")
    sleep: float = Field(0.01, description="set sleeptime between loop iterations")
    extended_addr: bool = Field(False, description="use extended isotp addresses")
    tester_addr: AutoInt = Field(0x6F1, description="tester address for --extended")
    query: bool = Field(False, description="query ECU description via RDBID")
    info_did: AutoInt = Field(0xF197, description="DID to query ECU description", metavar="DID")
    sniff_time: int = Field(
        5, description="Time in seconds to sniff on bus for current traffic", metavar="SECONDS"
    )

    @model_validator(mode="after")
    def check_transport_requirements(self) -> Self:
        if self.target is not None and (not self.target.scheme == RawCANTransport.SCHEME):
            raise ValueError(f"Unsupported transport schema {self.target.scheme}; must be can-raw!")
        if self.extended_addr and (self.start > 0xFF or self.stop > 0xFF):
            raise ValueError("start/stop maximum value is 0xFF")

        return self


class IsotpDiscoverer(UDSDiscoveryScanner):
    """Discovers all UDS endpoints on an ECU using ISO-TP normal addressing.
    This is the default protocol used by OBD.
    When using normal addressing, the ISO-TP header does not include an address and there is no generic tester address.
    Addressing is only done via CAN IDs. Every endpoint has a source and destination CAN ID.
    Typically, there is also a broadcast destination ID to address all endpoints."""

    CONFIG_TYPE = IsotpDiscovererConfig
    SHORT_HELP = "ISO-TP enumeration scanner"

    def __init__(self, config: IsotpDiscovererConfig):
        super().__init__(config)
        self.config: IsotpDiscovererConfig = config

    async def query_description(self, target_list: list[TargetURI], did: int) -> None:
        logger.info("reading info DID from all discovered endpoints")
        for target in target_list:
            logger.result("----------------------------")
            logger.result(f"Probing ECU: {target}")

            transport = await ISOTPTransport.connect(target)
            uds_client = UDSClient(transport, timeout=2)
            logger.result(f"reading device description at {g_repr(did)}")
            try:
                resp = await uds_client.read_data_by_identifier(did)
                if isinstance(resp, NegativeResponse):
                    logger.result(f"could not read did: {resp}")
                else:
                    logger.result(f"response was: {resp}")
            except Exception as e:
                logger.result(f"reading description failed: {e!r}")

    def _build_isotp_frame_extended(self, pdu: bytes, ext_addr: int) -> bytes:
        isotp_hdr = bytes([ext_addr, len(pdu) & 0x0F])
        return isotp_hdr + pdu

    def _build_isotp_frame(self, pdu: bytes) -> bytes:
        isotp_hdr = bytes([len(pdu) & 0x0F])
        return isotp_hdr + pdu

    def build_isotp_frame(
        self, req: UDSRequest, ext_addr: int | None = None, padding: int | None = None
    ) -> bytes:
        pdu = req.pdu
        max_pdu_len = 7 if ext_addr is None else 6
        if len(pdu) > max_pdu_len:
            raise ValueError("UDSRequest too large, ConsecutiveFrames not implemented")

        if ext_addr is None:
            frame = self._build_isotp_frame(pdu)
        else:
            frame = self._build_isotp_frame_extended(pdu, ext_addr)

        if padding is not None:
            pad_len = 8 - len(frame)
            frame += bytes([padding]) * pad_len

        return frame

    async def main(self) -> None:
        transport = await RawCANTransport.connect(self.config.target)
        found = []

        sniff_time: int = self.config.sniff_time
        logger.result(f"Recording idle bus communication for {sniff_time}s")
        addr_idle = await transport.get_idle_traffic(sniff_time)

        logger.result(f"Found {len(addr_idle)} CAN Addresses on idle Bus")
        transport.set_filter(addr_idle, inv_filter=True)

        req = UDSRequest.parse_dynamic(self.config.pdu)
        pdu = self.build_isotp_frame(req, padding=self.config.padding)

        for ID in range(self.config.start, self.config.stop + 1):
            await asyncio.sleep(self.config.sleep)

            dst_addr = self.config.tester_addr if self.config.extended_addr else ID
            if self.config.extended_addr:
                pdu = self.build_isotp_frame(req, ID, padding=self.config.padding)

            logger.info(f"Testing ID {can_id_repr(ID)}")
            is_broadcast = False

            await transport.sendto(pdu, timeout=0.1, dst=dst_addr)
            try:
                addr, payload = await transport.recvfrom(timeout=0.1)
                if addr == ID:
                    logger.info(f"The same CAN ID {can_id_repr(ID)} answered. Skipping…")
                    continue
            except TimeoutError:
                continue

            while True:
                # The recv buffer needs to be flushed to avoid
                # wrong results...
                try:
                    new_addr, _ = await transport.recvfrom(timeout=0.1)
                    if new_addr != addr:
                        is_broadcast = True
                        logger.result(
                            f"seems that broadcast was triggered on CAN ID {can_id_repr(ID)}, got answer from {can_id_repr(new_addr)}"
                        )
                    else:
                        logger.info(
                            f"seems like a large ISO-TP packet was received on CAN ID {can_id_repr(ID)}"
                        )
                except TimeoutError:
                    if is_broadcast:
                        logger.result(
                            f"seems that broadcast was triggered on CAN ID {can_id_repr(ID)}, got answer from {can_id_repr(addr)}"
                        )
                    else:
                        logger.result(
                            f"found endpoint on CAN ID [src:dst]: {can_id_repr(ID)}:{can_id_repr(addr)}: {payload.hex()}"
                        )
                        target_args = {}
                        target_args["is_fd"] = str(transport.config.is_fd).lower()
                        target_args["is_extended"] = str(transport.config.is_extended).lower()

                        if self.config.extended_addr:
                            target_args["ext_address"] = hex(ID)
                            target_args["rx_ext_address"] = hex(self.config.tester_addr & 0xFF)
                            target_args["src_addr"] = hex(self.config.tester_addr)
                            target_args["dst_addr"] = hex(addr)
                        else:
                            target_args["src_addr"] = hex(ID)
                            target_args["dst_addr"] = hex(addr)

                        if self.config.padding is not None:
                            target_args["tx_padding"] = f"{self.config.padding}"
                        if self.config.padding is not None:
                            target_args["rx_padding"] = f"{self.config.padding}"

                        hostname = self.config.target.hostname
                        assert hostname is not None

                        target = TargetURI.from_parts(
                            ISOTPTransport.SCHEME, hostname, None, target_args
                        )
                        found.append(target)
                    break

        logger.result(f"finished; found {len(found)} UDS endpoints")

        if self.artifacts_dir is not None:
            ecus_file = self.artifacts_dir.joinpath("ECUs.txt")
            logger.result(f"Writing urls to file: {ecus_file}")
            with ecus_file.open("w") as f:
                for target in found:
                    f.write(f"{target}\n")

        if self.db_handler is not None:
            logger.result("Writing urls to database")
            for target in found:
                await self.db_handler.insert_discovery_result(str(target))

        if self.config.query:
            await self.query_description(found, self.config.info_did)



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Parses Nested Pydantic Model Fields to Sub-Commands.

The `command` module contains the `should_parse` function, which checks whether
this module should be used to parse the field, as well as the `parse_field`
function, which parses nested `pydantic` model fields to `ArgumentParser`
sub-commands.
"""

import argparse
from typing import Any

from gallia.pydantic_argparse.utils.pydantic import (
    PydanticField,
)


def parse_field(
    subparser: argparse._SubParsersAction,  # type: ignore
    field: PydanticField,
    extra_defaults: dict[type, dict[str, Any]] | None = None,
) -> None:
    """Adds command pydantic field to argument parser.

    Args:
        subparser (argparse._SubParsersAction): Sub-parser to add to.
        field (PydanticField): Field to be added to parser.
        extra_defaults: Defaults coming from external sources, such as environment variables or config files.
    """
    # Add Command
    subparser.add_parser(
        field.info.title or field.info.alias or field.name,
        help=field.info.description,
        model=field.model_type,
        exit_on_error=False,  # Allow top level parser to handle exiting
        extra_defaults=extra_defaults,
    )



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import socket
import sys
from collections.abc import Iterable
from itertools import product
from urllib.parse import parse_qs, urlparse

from gallia.command import AsyncScript
from gallia.command.base import AsyncScriptConfig
from gallia.command.config import AutoInt, Field
from gallia.log import get_logger
from gallia.net import net_if_broadcast_addrs
from gallia.services.uds.core.service import TesterPresentRequest, TesterPresentResponse
from gallia.transports.doip import (
    DiagnosticMessageNegativeAckCodes,
    DoIPConnection,
    DoIPEntityStatusResponse,
    DoIPNegativeAckError,
    DoIPRoutingActivationDeniedError,
    GenericHeader,
    PayloadTypes,
    ProtocolVersions,
    RoutingActivationRequestTypes,
    RoutingActivationResponseCodes,
    TimingAndCommunicationParameters,
    VehicleAnnouncementMessage,
)

logger = get_logger(__name__)


class DoIPDiscovererConfig(AsyncScriptConfig):
    start: AutoInt = Field(
        0x00, description="Set start address of TargetAddress search range", metavar="INT"
    )
    stop: AutoInt = Field(
        0xFFFF, description="Set stop address of TargetAddress search range", metavar="INT"
    )
    target: str | None = Field(
        None,
        description="The more you give, the more automatic detection will be skipped: IP, Port, RoutingActivationType, SourceAddress",
        metavar="<DoIP URL target string>",
    )
    timeout: float | None = Field(
        None,
        description="This flag overrides the default timeout of DiagnosticMessages, which can be used to fine-tune classification of unresponsive ECUs or broadcast detection",
        metavar="SECONDS (FLOAT)",
    )
    tcp_connect_delay: float = Field(
        0.0,
        description="This flag delays subsequent TCP connect attempts during all enumerations. Useful if the DoIP entity requires some time before accepting new TCP requests.",
        metavar="SECONDS (FLOAT)",
    )


class DoIPDiscoverer(AsyncScript):
    """This script scans for active DoIP endpoints and automatically enumerates allowed
    RoutingActivationTypes and known SourceAddresses. Once valid endpoints are acquired,
    the script continues to discover valid TargetAddresses that are accepted and respond
    to UDS TesterPresent requests."""

    CONFIG_TYPE = DoIPDiscovererConfig
    SHORT_HELP = "zero-knowledge DoIP enumeration scanner"

    protocol_version = ProtocolVersions.ISO_13400_2_2019.value

    def __init__(self, config: DoIPDiscovererConfig):
        super().__init__(config)
        self.config: DoIPDiscovererConfig = config

    async def main(self) -> None:
        logger.notice("[👋] Welcome to @realDoIP-Discovery powered by MoarMemes…")

        target = urlparse(self.config.target) if self.config.target is not None else None
        if target is not None and target.scheme != "doip":
            logger.error("[🫣] --target must be doip://…")
            sys.exit(2)

        if self.db_handler is not None:
            try:
                await self.db_handler.insert_discovery_run("doip")
            except Exception as e:
                logger.warning(f"Could not write the discovery run to the database: {e!r}")

        # Set TCP connect delay for RoutingActivationType and SourceAddress enumeration
        tcp_connect_delay: float = self.config.tcp_connect_delay

        # Discover Hostname and Port
        tgt_hostname: str
        tgt_port: int
        if target is not None and target.hostname is not None and (target.port is not None):
            logger.notice("[📋] Skipping host/port discovery because given by --target")
            tgt_hostname = target.hostname
            tgt_port = target.port
        else:
            logger.notice("[🔍] Discovering Host and Port via UDP Broadcast")

            hosts = await self.run_udp_discovery()

            if len(hosts) != 1:
                logger.error("[🍃] Can only continue with a single DoIP host! Give me a --target!")
                sys.exit(11)

            tgt_hostname, tgt_port = hosts[0]

        # Politely ask for more details via UDP
        await self.gather_doip_details(tgt_hostname, tgt_port)

        # Enumerate all valid RoutingActivationType/Source Address tuples, but only if required
        rat_not_unsupported: Iterable[int]
        rat_not_unknown: Iterable[int]
        if target is not None and "activation_type" in parse_qs(target.query):
            logger.notice("[📋] Skipping RoutingActivationType discovery because given by --target")
            rat_not_unsupported = [int(parse_qs(target.query)["activation_type"][0], 0)]
        else:
            rat_not_unsupported = range(0x100)
        if target is not None and "src_addr" in parse_qs(target.query):
            logger.notice("[📋] Skipping SourceAddress discovery because given by --target")
            rat_not_unknown = [int(parse_qs(target.query)["src_addr"][0], 0)]
        else:
            rat_not_unknown = range(0x10000)

        # We need to know whether the DoIP entity checks for RoutingActivationTypes or SourceAddresses first
        # By requesting a RoutingActivation with reserved RAT and reserved SrcAddr, we see it based on the error
        a, _, _ = await self.enumerate_routing_activation_requests(
            tgt_hostname, tgt_port, [0x02], [0x00], tcp_connect_delay
        )
        routing_activation_types_first = len(a) == 0

        if routing_activation_types_first is True and len(rat_not_unsupported) != 1:
            logger.notice("[🔍] Enumerating RoutingActivationTypes")
            rat_not_unsupported, _, _ = await self.enumerate_routing_activation_requests(
                tgt_hostname, tgt_port, range(0x100), [0x00], tcp_connect_delay
            )
            logger.notice(
                f"[💎] Look what promising RoutingActivationTypes I've found: {', '.join([f'{x:#x}' for x in rat_not_unsupported])}"
            )
        elif routing_activation_types_first is False and len(rat_not_unknown) != 1:
            logger.notice("[🔍] Enumerating SourceAddresses")
            _, rat_not_unknown, _ = await self.enumerate_routing_activation_requests(
                tgt_hostname, tgt_port, [0x02], range(0x10000), tcp_connect_delay
            )
            logger.notice(
                f"[💎] Look what promising SourceAddresses I've found: {', '.join([f'{x:#x}' for x in rat_not_unknown])}"
            )

        logger.notice("[🔍] Enumerating valid RoutingActivationType/SourceAddress tuples")
        _, _, targets = await self.enumerate_routing_activation_requests(
            tgt_hostname, tgt_port, rat_not_unsupported, rat_not_unknown, tcp_connect_delay
        )

        if len(targets) != 1:
            logger.error(
                f"[💣] I found {len(targets)} valid RoutingActivationType/SourceAddress tuples, but can only continue with exactly one; choose your weapon with --target!"
            )
            sys.exit(20)

        # Enumerate valid TargetAddresses
        if target is not None and "target_addr" in parse_qs(target.query):
            logger.error(
                "[😵] Why do you give me a target_addr in --target? Am I useless to you??? GOODBYE!"
            )
            sys.exit(3)

        logger.notice(
            f"[🔍] Enumerating all TargetAddresses from {self.config.start:#x} to {self.config.stop:#x}"
        )

        target = urlparse(targets[0])
        tgt_src = int(parse_qs(target.query)["src_addr"][0], 0)
        tgt_rat = int(parse_qs(target.query)["activation_type"][0], 0)

        await self.enumerate_target_addresses(
            tgt_hostname,
            tgt_port,
            tgt_rat,
            tgt_src,
            self.config.start,
            self.config.stop,
            tcp_connect_delay,
            self.config.timeout,
        )

        logger.notice("[🛩️] All done, thanks for flying with us!")

    async def gather_doip_details(self, tgt_hostname: str, tgt_port: int) -> None:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setblocking(False)
        sock.bind(("0.0.0.0", 0))
        loop = asyncio.get_running_loop()

        # We send two packets: a VehicleIdentificationRequest and a DoIPEntityStatusRequest that
        # only differ in their response, so let's reuse code...
        for req_type in [
            PayloadTypes.VehicleIdentificationRequestMessage,
            PayloadTypes.DoIPEntityStatusRequest,
        ]:
            logger.info(f"[🥚] Sending {req_type.name}...")

            hdr = GenericHeader(
                ProtocolVersion=self.protocol_version
                if req_type == PayloadTypes.DoIPEntityStatusRequest
                else 0xFF,
                PayloadType=req_type,
                PayloadLength=0,
            )
            await loop.sock_sendto(sock, hdr.pack(), (tgt_hostname, tgt_port))

            try:
                data, _ = await asyncio.wait_for(loop.sock_recvfrom(sock, 1024), 2)
            except TimeoutError:
                logger.info("[🐣] No response!")
                continue

            hdr = GenericHeader.unpack(data[:8])

            if hdr.PayloadType == PayloadTypes.VehicleAnnouncementMessage:
                logger.notice(
                    f"[👮] Identification please: {VehicleAnnouncementMessage.unpack(data[8:])}"
                )
                logger.info(f"[🎯] Setting protocol version to {hdr.ProtocolVersion}")
                self.protocol_version = hdr.ProtocolVersion
            elif hdr.PayloadType == PayloadTypes.DoIPEntityStatusResponse:
                status = DoIPEntityStatusResponse.unpack(data[8:])
                logger.notice(
                    f"[👏] This DoIP entity is a {status.NodeType.name} with {status.CurrentlyOpenTCP_DATASockets}/{status.MaximumConcurrentTCP_DATASockets} concurrent TCP sockets currently open and a maximum data size of {status.MaximumDataSize} ({status.MaximumDataSize:#0x})."
                )

        sock.close()

    async def enumerate_routing_activation_requests(
        self,
        tgt_hostname: str,
        tgt_port: int,
        routing_activation_types: Iterable[int],
        source_addresses: Iterable[int],
        tcp_connect_delay: float,
    ) -> tuple[list[int], list[int], list[str]]:  # noqa: PLR0913
        rat_not_unsupported: list[int] = []
        rat_not_unknown: list[int] = []
        targets: list[str] = []

        for routing_activation_type, source_address in product(
            routing_activation_types, source_addresses
        ):
            try:  # Dummy target address, never actually used
                # Ensure that connections do not remain in TIME_WAIT
                conn = await DoIPConnection.connect(
                    tgt_hostname,
                    tgt_port,
                    source_address,
                    0xAFFE,
                    so_linger=True,
                    protocol_version=self.protocol_version,
                )
            except OSError as e:
                logger.warning(
                    f"[🚨] J.A.R.V.I.S., recheck rat {routing_activation_type:#x} and src_addr {source_address:#x}; they failed with {e!r}"
                )
                continue

            try:
                await conn.write_routing_activation_request(routing_activation_type)
            except DoIPRoutingActivationDeniedError as e:
                logger.info(
                    f"[🌟] Brilliant, RoutingActivationType {routing_activation_type:#x} and SourceAddress {source_address:#x} yields {e.rac_code.name}"
                )

                if e.rac_code != RoutingActivationResponseCodes.UnsupportedActivationType:
                    rat_not_unsupported.append(routing_activation_type)
                if e.rac_code != RoutingActivationResponseCodes.UnknownSourceAddress:
                    rat_not_unknown.append(source_address)
                continue
            finally:
                await conn.close()
                await asyncio.sleep(tcp_connect_delay)

            targets.append(
                f"doip://{tgt_hostname}:{tgt_port}?protocol_version={self.protocol_version}&activation_type={routing_activation_type:#x}&src_addr={source_address:#x}"
            )
            logger.notice(f"[🤯] Holy moly, it actually worked: {targets[-1]}")

            if self.artifacts_dir is not None:
                with self.artifacts_dir.joinpath("1_valid_routing_activation_requests.txt").open(
                    "a"
                ) as f:
                    f.write(f"{targets[-1]}\n")

        if len(targets) > 0:
            logger.notice("[⚔️] It's dangerous to test alone, take one of these:")
            for item in targets:
                logger.notice(item)

        return (rat_not_unsupported, rat_not_unknown, targets)

    async def enumerate_target_addresses(
        self,
        tgt_hostname: str,
        tgt_port: int,
        correct_rat: int,
        correct_src: int,
        start: int,
        stop: int,
        tcp_connect_delay: float,
        timeout: None | float = None,
    ) -> None:  # noqa: PLR0913
        known_targets = []
        unreachable_targets = []
        search_space = range(start, stop + 1)

        target_template = f"doip://{tgt_hostname}:{tgt_port}?protocol_version={self.protocol_version}&activation_type={correct_rat:#x}&src_addr={correct_src:#x}&target_addr={{:#x}}"
        conn = await self.create_DoIP_conn(
            tgt_hostname, tgt_port, correct_rat, correct_src, 0xAFFE, fast_queue=True
        )
        reader_task = asyncio.create_task(self.task_read_diagnostic_messages(conn, target_template))

        for target_addr in search_space:
            logger.debug(f"[🚧] Attempting connection to {target_addr:#x}")

            conn.target_addr = target_addr
            current_target = target_template.format(target_addr)

            try:
                req = TesterPresentRequest(suppress_response=False)
                await conn.write_diag_request(req.pdu)

                # If we reach this, the request was not denied due to unknown TargetAddress or other DoIP errors
                known_targets.append(current_target)
                logger.notice(f"[🥈] HEUREKA: target address {target_addr:#x} is valid! ")
                if self.artifacts_dir is not None:
                    with self.artifacts_dir.joinpath("3_valid_targets.txt").open("a") as f:
                        f.write(f"{current_target}\n")

                # Here is where "reader_task" comes into play, which monitors incoming DiagnosticMessage replies

            except DoIPNegativeAckError as e:
                if e.nack_code == DiagnosticMessageNegativeAckCodes.UnknownTargetAddress:
                    logger.info(f"[🫥] {target_addr:#x} is an unknown target address")
                    continue
                elif e.nack_code == DiagnosticMessageNegativeAckCodes.TargetUnreachable:
                    logger.info(f"[💤] {target_addr:#x} is (currently?) unreachable")
                    unreachable_targets.append(current_target)
                    if self.artifacts_dir is not None:
                        with self.artifacts_dir.joinpath("5_unreachable_targets.txt").open(
                            "a"
                        ) as f:
                            f.write(f"{current_target}\n")
                    continue
                else:
                    logger.warning(
                        f"[🤷] {target_addr:#x} is behaving strangely: {e.nack_code.name}"
                    )
                    if self.artifacts_dir is not None:
                        with self.artifacts_dir.joinpath("7_targets_with_errors.txt").open(
                            "a"
                        ) as f:
                            f.write(f"{target_addr:#x}: {e.nack_code.name}\n")
                    continue

            except ConnectionError as e:
                # Whenever this triggers, but sometimes connections are closed not by us
                logger.warning(f"[🫦] Sexy, but unexpected: {target_addr:#x} triggered {e!r}")
                if self.artifacts_dir is not None:
                    with self.artifacts_dir.joinpath("7_targets_with_errors.txt").open("a") as f:
                        f.write(f"{target_addr:#x}: {e}\n")
                # Re-establish DoIP connection
                await conn.close()
                await asyncio.sleep(tcp_connect_delay)

                conn = await self.create_DoIP_conn(
                    tgt_hostname, tgt_port, correct_rat, correct_src, 0xAFFE
                )
                continue

        logger.notice(
            f"[⚔️] It's dangerous to test alone, take one of these {len(known_targets)} known targets:"
        )
        if len(known_targets) > 100:
            logger.notice("[💯] Too many to print, check the artifacts instead!")
        else:
            for item in known_targets:
                logger.notice(item)

        logger.notice(
            f"[❓] Those {len(unreachable_targets)} targets were unreachable by the gateway (could be just temporary):"
        )
        for item in unreachable_targets:
            logger.notice(item)

        logger.info("[😴] Giving all ECUs a chance to reply...")
        await asyncio.sleep(TimingAndCommunicationParameters.DiagnosticMessageMessageTimeout / 1000)
        reader_task.cancel()
        await reader_task
        await conn.close()

        logger.notice(
            f"[🧭] Check out the content of the log files at {self.artifacts_dir} as well!"
        )

    async def task_read_diagnostic_messages(
        self, conn: DoIPConnection, target_template: str
    ) -> None:
        responsive_targets = []
        potential_broadcasts = []
        try:
            while True:
                _, payload = await conn.read_diag_request_raw()
                (source_address, data) = (payload.SourceAddress, payload.UserData)
                current_target = target_template.format(source_address)

                resp = TesterPresentResponse.parse_static(data)
                logger.notice(f"[🥇] It cannot get nicer: {source_address:#x} responded: {resp}")

                if current_target not in responsive_targets:
                    responsive_targets.append(current_target)
                    if self.artifacts_dir is not None:
                        with self.artifacts_dir.joinpath("4_responsive_targets.txt").open("a") as f:
                            f.write(f"{current_target}\n")
                    if self.db_handler is not None:
                        await self.db_handler.insert_discovery_result(current_target)

                if (
                    abs(source_address - conn.target_addr) > 10
                    and conn.target_addr not in potential_broadcasts
                ):
                    potential_broadcasts.append(conn.target_addr)

        except asyncio.CancelledError:
            logger.debug("Diagnostic Message reader got cancelled")
        except Exception as e:
            logger.error(f"Diagnostic Message reader died with {e!r}")

        finally:
            logger.notice(
                f"[💰] For even more profit, try one of the {len(responsive_targets)} targets that actually responded:"
            )
            for item in responsive_targets:
                logger.notice(item)

            # TODO: the discoverer could be extended to search for and validate the broadcast address(es) automatically
            if len(potential_broadcasts) > 0:
                logger.notice(
                    "[🕵️] You could also investigate these target addresses that appear to be near broadcasts:"
                )
            for target_addr in potential_broadcasts:
                logger.notice(f"[🤑] B-B-B-B-B-B-BROADCAST around TargetAddress {target_addr:#x}!")

    async def create_DoIP_conn(
        self,
        hostname: str,
        port: int,
        routing_activation_type: int,
        src_addr: int,
        target_addr: int,
        fast_queue: bool = False,
    ) -> DoIPConnection:  # noqa: PLR0913
        while True:
            try:  # Ensure that connections do not remain in TIME_WAIT
                conn = await DoIPConnection.connect(
                    hostname,
                    port,
                    src_addr,
                    target_addr,
                    so_linger=True,
                    protocol_version=self.protocol_version,
                    separate_diagnostic_message_queue=fast_queue,
                )
                logger.info("[📫] Sending RoutingActivationRequest")
                await conn.write_routing_activation_request(
                    RoutingActivationRequestTypes(routing_activation_type)
                )
            except Exception as e:  # TODO: this probably is too broad
                logger.warning(
                    f"[🫨] Got me some good errors when it should be working (dis an infinite loop): {e!r}"
                )
                continue
            return conn

    async def run_udp_discovery(self) -> list[tuple[str, int]]:
        addrs = net_if_broadcast_addrs()
        found = []

        for addr in addrs:
            logger.info(f"[💌] Sending DoIP VehicleIdentificationRequest to {addr.broadcast}")
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            sock.setblocking(False)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            sock.bind((str(addr.local), 0))
            loop = asyncio.get_running_loop()

            hdr = GenericHeader(0xFF, PayloadTypes.VehicleIdentificationRequestMessage, 0x00)
            await loop.sock_sendto(sock, hdr.pack(), (str(addr.broadcast), 13400))
            try:
                while True:
                    data, from_addr = await asyncio.wait_for(loop.sock_recvfrom(sock, 1024), 2)
                    info = VehicleAnnouncementMessage.unpack(data[8:])
                    logger.notice(f"[💝]: {from_addr} responded: {info}")
                    found.append(from_addr)
            except TimeoutError:
                logger.info("[💔] Reached timeout...")
                continue
            finally:
                sock.close()

        if len(found) > 0:
            logger.notice("[💎] Look what valid hosts I've found:")
            for item in found:
                url = f"doip://{item[0]}:{item[1]}"
                logger.notice(url)
                if self.artifacts_dir is not None:
                    with self.artifacts_dir.joinpath("0_valid_hosts.txt").open("a") as f:
                        f.write(f"{url}\n")
        else:
            logger.notice(
                "[👸] Your princess is in another castle: no DoIP endpoints here it seems..."
            )

        return found



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Types Utility Functions for Declarative Typed Argument Parsing.

The `types` module contains a utility function used for determining and
comparing the types of `pydantic fields.
"""

from collections.abc import Iterable
from typing import Any


def all_types(types: Iterable[Any]) -> bool:
    """Check if all inputs are `type`s and not instances.

    Args:
        types (Iterable): an interable of putative `type` objects

    Returns:
        bool: whether or not all inputs are `type`s
    """
    return all(isinstance(t, type) for t in types)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

from abc import ABC, abstractmethod
from collections.abc import Mapping, MutableMapping
from dataclasses import dataclass
from importlib.metadata import entry_points
from typing import Union

from gallia.command import BaseCommand
from gallia.services.uds import ECU
from gallia.transports import BaseTransport, TargetURI


@dataclass
class CommandTree:
    description: str | None
    subtree: MutableMapping[str, Union["CommandTree", type[BaseCommand]]]


class Plugin(ABC):
    @classmethod
    @abstractmethod
    def name(cls) -> str: ...

    @classmethod
    def description(cls) -> str:
        return ""

    @classmethod
    def transports(cls) -> list[type[BaseTransport]]:
        return []

    @classmethod
    def ecus(cls) -> list[type[ECU]]:
        return []

    @classmethod
    def commands(cls) -> Mapping[str, CommandTree | type[BaseCommand]]:
        return {}


def load_plugins() -> list[type[Plugin]]:
    """Loads the ``gallia_transports`` entry_point."""
    plugins: list[type[Plugin]] = []

    for plugin_ep in entry_points(group="gallia_plugins"):
        plugin = plugin_ep.load()

        if not issubclass(plugin, Plugin):
            raise ValueError(
                f"{plugin.__name__} from {plugin_ep.name} is not derived from {Plugin.__name__}"
            )

        plugins.append(plugin)

    return plugins


def load_transports() -> list[type[BaseTransport]]:
    transports = []

    for plugin in load_plugins():
        for transport in plugin.transports():
            transports.append(transport)

    return transports


def load_transport(target: TargetURI) -> type[BaseTransport]:
    """Selects a transport class depending on a TargetURI.
    The lookup is performed in builtin transports and all
    classes behind the ``gallia_transports`` entry_point.
    """
    for plugin in load_plugins():
        for transport in plugin.transports():
            if target.scheme == transport.SCHEME:
                return transport

    raise ValueError(f"no transport for {target}")


def load_ecus() -> list[type[ECU]]:
    ecus = []

    for plugin in load_plugins():
        for ecu in plugin.ecus():
            ecus.append(ecu)

    return ecus


def load_ecu(vendor: str) -> type[ECU]:
    """Selects an ecu class depending on a vendor string.
    The lookup is performed in builtin ecus and all
    classes behind the ``gallia_uds_ecus`` entry_point.
    The vendor string ``default`` selects a generic ECU.
    """
    for plugin in load_plugins():
        for ecu in plugin.ecus():
            if vendor == ecu.OEM:
                return ecu

    raise ValueError(f"no such OEM: '{vendor}'")


def _merge_commands(
    c1: MutableMapping[str, CommandTree | type[BaseCommand]],
    c2: Mapping[str, CommandTree | type[BaseCommand]],
) -> None:
    for key, value in c2.items():
        if key not in c1:
            c1[key] = value
        elif isinstance(value, CommandTree) and isinstance(cmd := c1[key], CommandTree):
            try:
                _merge_command_trees(cmd, value)
            except ValueError as e:
                raise ValueError(f"{key} {str(e)}")
        else:
            raise ValueError(f"{key} ]: There already exists a leaf command")


def _merge_command_trees(tree1: CommandTree, tree2: CommandTree) -> None:
    if (
        tree1.description is not None
        and tree2.description is not None
        and tree1.description != tree2.description
    ):
        raise ValueError("]: Incompatible descriptions")

    _merge_commands(tree1.subtree, tree2.subtree)


def load_commands() -> MutableMapping[str, CommandTree | type[BaseCommand]]:
    plugins = load_plugins()
    commands: MutableMapping[str, CommandTree | type[BaseCommand]] = {}

    for plugin in plugins:
        try:
            _merge_commands(commands, plugin.commands())
        except ValueError as e:
            raise ValueError(
                f'Plugin "{plugin.name()}" conflicts with other plugins on command [ {str(e)}'
            ) from None

    return commands



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0


from gallia.command import UDSScanner
from gallia.command.config import Field
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds.core.service import NegativeResponse

logger = get_logger(__name__)


class VINPrimitiveConfig(UDSScannerConfig):
    properties: bool = Field(
        False,
        description="Read and store the ECU proporties prior and after scan",
        cli_group=UDSScannerConfig._cli_group,
        config_section=UDSScannerConfig._config_section,
    )


class VINPrimitive(UDSScanner):
    """Request VIN"""

    CONFIG_TYPE = VINPrimitiveConfig
    SHORT_HELP = "request VIN"

    def __init__(self, config: VINPrimitiveConfig):
        super().__init__(config)
        self.config: VINPrimitiveConfig = config

    async def main(self) -> None:
        resp = await self.ecu.read_vin()
        if isinstance(resp, NegativeResponse):
            logger.warning(f"ECU said: {resp}")
            return
        logger.result(resp.data_record.hex())



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Recursively Nesting Sub-Parsers Action for Typed Argument Parsing.

The `actions` module contains the `SubParsersAction` class, which is an action
that provides recursive namespace nesting when parsing sub-commands. It also
contains the `BooleanOptionalAction` class, which is a direct backport of the
Python standard library `argparse` class of the same name.
"""

import argparse
from collections.abc import Callable, Iterable, Sequence
from typing import (
    Any,
    TypeVar,
    cast,
)


class SubParsersAction(argparse._SubParsersAction):  # type: ignore
    """Recursively Nesting Sub-Parsers Action for Typed Argument Parsing.

    This custom action differs in functionality from the existing standard
    argparse SubParsersAction because it nests the resultant sub-namespace
    directly into the supplied parent namespace, rather than iterating through
    and updating the parent namespace object with each argument individually.

    Example:
        Construct `ArgumentParser`:
        ```python
        # Create Argument Parser
        parser = argparse.ArgumentParser()

        # Add Example Global Argument
        parser.add_argument("--time")

        # Add SubParsersAction
        subparsers = parser.add_subparsers()

        # Add Example 'walk' Command with Arguments
        walk = subparsers.add_parser("walk")
        walk.add_argument("--speed")
        walk.add_argument("--distance")

        # Add Example 'talk' Command with Arguments
        talk = subparsers.add_parser("talk")
        talk.add_argument("--volume")
        talk.add_argument("--topic")
        ```

        Parse the Arguments:
        ```console
        --time 3 walk --speed 7 --distance 42
        ```

        Check Resultant Namespaces:
        ```python
        Original: Namespace(time=3, speed=7, distance=42)
        Custom:   Namespace(time=3, walk=Namespace(speed=7, distance=42))
        ```

    This behaviour results in a final namespace structure which is much easier
    to parse, where subcommands are easily identified and nested into their own
    namespace recursively.
    """

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        """Parses arguments into a namespace with the specified subparser.

        This custom method parses arguments with the specified subparser, then
        embeds the resultant sub-namespace into the supplied parent namespace.

        Args:
            parser (argparse.ArgumentParser): Parent argument parser object.
            namespace (argparse.Namespace): Parent namespace being parsed to.
            values (Union[str, Sequence[Any], None]): Arguments to parse.
            option_string (Optional[str]): Optional option string (not used).

        Raises:
            argparse.ArgumentError: Raised if subparser name does not exist.
        """
        # Check values object is a sequence
        # In order to not violate the Liskov Substitution Principle (LSP), the
        # function signature for __call__ must match the base Action class. As
        # such, this function signature also accepts 'str' and 'None' types for
        # the values argument. However, in reality, this should only ever be a
        # list of strings here, so we just do a type cast.
        values = cast(list[str], values)

        # Get Parser Name and Remaining Argument Strings
        parser_name, *arg_strings = values

        # Try select the parser
        try:
            # Select the parser
            parser = self._name_parser_map[parser_name]

        except KeyError as exc:
            # Parser doesn't exist, raise an exception
            raise argparse.ArgumentError(
                self,
                f"unknown parser {parser_name} (choices: {', '.join(self._name_parser_map)})",
            ) from exc

        # Parse all the remaining options into a sub-namespace, then embed this
        # sub-namespace into the parent namespace
        subnamespace, arg_strings = parser.parse_known_args(arg_strings)
        setattr(namespace, parser_name, subnamespace)

        # Store any unrecognized options on the parent namespace, so that the
        # top level parser can decide what to do with them
        if arg_strings:
            vars(namespace).setdefault(argparse._UNRECOGNIZED_ARGS_ATTR, [])
            getattr(namespace, argparse._UNRECOGNIZED_ARGS_ATTR).extend(arg_strings)


T = TypeVar("T")


class BooleanOptionalAction(argparse.Action):  # pragma: no cover
    """Action for parsing paired GNU-style boolean arguments.

    This backported action provides the functionality for parsing paired
    GNU-style boolean arguments, such as "--foo/--no-foo". This style of
    argument allows us to easily provide *required* boolean arguments.

    This action was added into the Python standard library `argparse` module
    in [`BPO-8538`](https://bugs.python.org/issue8538) and is available in
    Python 3.9 and above. In order to support Python 3.7 and 3.8 we directly
    backport the class and make it available here.

    Source:
    <https://github.com/python/cpython/blob/v3.11.0/Lib/argparse.py#L878-L914>
    """

    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str,
        default: T | str | None = None,
        type_: Callable[[str], T] | argparse.FileType | None = None,
        choices: Iterable[T] | None = None,
        required: bool = False,
        help: str | None = None,  # noqa: A002
        metavar: str | tuple[str, ...] | None = None,
    ) -> None:
        """Instantiates the Boolean Optional Action.

        This creates the default provided "--<OPT>" option strings which set
        the argument to `True`. It also creates alternative pair "--no-<OPT>"
        option strings which set the argument to `False`.

        Args:
            option_strings (Sequence[str]): Option strings.
            dest (str): Destination variable to save the value to.
            default (Optional[Union[T, str]]): Default value of the option.
            type (Optional[Union[Callable[[str], T], argparse.FileType]]): Type
                to cast the option to.
            choices (Optional[Iterable[T]]): Allowed values for the option.
            required (bool): Whether the option is required.
            help (Optional[str]): Help string for the option.
            metavar (Optional[Union[str, Tuple[str, ...]]]): Meta variable name
                for the option.
        """
        # Initialise intermediary option strings list
        _option_strings = []

        # Loop through passed in option strings
        for option_string in option_strings:
            # Append the option string to the new list
            _option_strings.append(option_string)

            # Check if this option string is a "--<OPT>" option string
            if option_string.startswith("--"):
                # Create a "--no-<OPT>" negated option string
                _option_strings.append(f"--no-{option_string[2:]}")

        # Initialise Super Class
        super().__init__(
            option_strings=_option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type_,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        """Parses the provided boolean arguments into a namespace.

        This custom method parses arguments as booleans, negating the values of
        any arguments prepended with "--no-".

        Args:
            parser (argparse.ArgumentParser): Parent argument parser object.
            namespace (argparse.Namespace): Parent namespace being parsed to.
            values (Optional[Union[str, Sequence[Any]]]): Arguments to parse.
            option_string (Optional[str]): Optional option string.
        """
        # Check if the passed in option string matches our option strings
        if option_string in self.option_strings:
            # Set a boolean value on the namespace
            # If the option string starts with "--no-", then negate the value
            setattr(namespace, self.dest, not option_string.startswith("--no-"))  # type: ignore[union-attr]

    def format_usage(self) -> str:
        """Formats the usage string.

        Returns:
            str: Usage string for the option.
        """
        # Format and return usage string
        return " | ".join(self.option_strings)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import gzip
import io
import os
import shlex
import struct
import sys
from asyncio import subprocess
from datetime import datetime
from pathlib import Path
from socket import SocketKind
from typing import Self, cast
from urllib.parse import urlparse

from gallia.log import get_logger
from gallia.net import split_host_port
from gallia.transports import TargetURI, TransportScheme
from gallia.utils import auto_int, handle_task_error, set_task_handler_ctx_variable

logger = get_logger(__name__)


if sys.platform.startswith("linux") or sys.platform == "darwin":

    class Dumpcap:
        BUFSIZE = io.DEFAULT_BUFFER_SIZE

        def __init__(
            self,
            proc: subprocess.Process,
            artifacts_dir: Path,
            outfile: Path,
            cleanup: int = 2,
        ) -> None:
            self.proc = proc
            self.artifacts_dir = artifacts_dir
            self.outfile = outfile
            self.cleanup = cleanup
            self.ready_event = asyncio.Event()
            self.compressor = asyncio.create_task(self._compressor())
            self.compressor.add_done_callback(
                handle_task_error,
                context=set_task_handler_ctx_variable(__name__, "Dumpcap"),
            )

        @classmethod
        async def start(
            cls,
            target: TargetURI,
            artifacts_dir: Path,
        ) -> Self | None:
            ts = int(datetime.now().timestamp())

            match target.scheme:
                case TransportScheme.ISOTP | TransportScheme.CAN_RAW:
                    outfile = artifacts_dir.joinpath(f"candump-{ts}.pcap.gz")
                    src_addr = (
                        auto_int(target.qs["src_addr"][0]) if "src_addr" in target.qs else None
                    )
                    dst_addr = (
                        auto_int(target.qs["dst_addr"][0]) if "dst_addr" in target.qs else None
                    )
                    cmd = cls._can_cmd(
                        target.netloc,
                        src_addr,
                        dst_addr,
                    )
                case TransportScheme.UNIX | TransportScheme.UNIX_LINES:
                    logger.warning("Dumpcap does not support unix domain sockets")
                    return None
                # There is currently no API for transport plugins to
                # register a scheme and a corresponding invocation
                # for dumpcap. So this match…case is best effort,
                # since it defaults to ethernet.
                case _:
                    outfile = artifacts_dir.joinpath(f"eth-{ts}.pcap.gz")
                    cmd = await cls._eth_cmd(target.netloc)

            if cmd is None:
                return None

            cmd_str = shlex.join(cmd)
            try:
                proc = await asyncio.create_subprocess_exec(
                    *cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.DEVNULL,
                )
                await asyncio.sleep(0.2)
            except Exception as e:
                logger.error(f"Could not start dumpcap: ({e!r})")
                return None

            if proc.returncode:
                logger.error(f"dumpcap terminated with exit code: [{proc.returncode}]")
                return None

            logger.info(f'Started "dumpcap": {cmd_str}')

            return cls(proc, artifacts_dir, outfile)

        async def sync(self, timeout: float = 1) -> None:
            await asyncio.wait_for(self.ready_event.wait(), timeout)

        async def stop(self) -> None:
            logger.info(f"Waiting {self.cleanup}s for dumpcap to receive all packets")
            await asyncio.sleep(self.cleanup)
            try:
                self.proc.terminate()
            except ProcessLookupError:
                logger.warning("dumpcap terminated before gallia")
            await self.proc.wait()
            await self.compressor

        async def _compressor(self) -> None:
            ready = False
            assert self.proc.stdout
            with await asyncio.to_thread(gzip.open, self.outfile, "wb") as f:
                while True:
                    chunk = await self.proc.stdout.read(64 * 1024)
                    if chunk == b"":
                        break
                    # Dumpcap first writes the pcap header. It does this
                    # once the tool is ready. We can use this is as a poor
                    # man's synchronization primitive.
                    if not ready:
                        ready = True
                        self.ready_event.set()
                    await asyncio.to_thread(f.write, chunk)

        @staticmethod
        def _swap_bytes_16(x: int) -> int:
            return cast(int, struct.unpack(">H", struct.pack("<H", x))[0])

        @staticmethod
        def _can_cmd(iface: str, src_addr: int | None, dst_addr: int | None) -> list[str] | None:
            args = ["dumpcap", "-q", "-i", iface, "-w", "-"]
            # Debug this with `dumpcap -d` or `tshark -x` to inspect the captured buffer.
            filter_ = ""

            if src_addr is not None and dst_addr is not None:
                # TODO: Support extended CAN IDs
                if src_addr > 0x800 or dst_addr > 0x800:
                    logger.error("Extended CAN Ids are currently not supported!")
                    return None

                filter_ += (
                    f"link[0:2] == {Dumpcap._swap_bytes_16(src_addr):#x} "  # can_id is in little endian
                    f"|| link[0:2] == {Dumpcap._swap_bytes_16(dst_addr):#x}"
                )
            args += ["-f", filter_]
            return args

        @staticmethod
        async def _eth_cmd(target_ip: str) -> list[str] | None:
            try:
                host, port = split_host_port(target_ip)
            except Exception as e:
                logger.error(f"Invalid argument for target ip: {target_ip}; {e}")
                return None

            if proxy := os.getenv("all_proxy"):
                url = urlparse(proxy)
                host = str(url.hostname) if url.hostname else "localhost"
                port = url.port if url.port else 1080

            loop = asyncio.get_running_loop()
            res = await loop.getaddrinfo(
                host,
                port,
                type=SocketKind.SOCK_STREAM,
            )
            # We don't do round robin, ergo we only have
            # 1 result. That's fine here. The data structure
            # of the return value of getaddrinfo() is weird,
            # but it's documented in the python docs:
            # https://docs.python.org/3/library/socket.html
            addr_tuple = res[0]
            ip, port = addr_tuple[4][0], addr_tuple[4][1]
            return [
                "dumpcap",
                "-q",
                "-i",
                "any",
                "-w",
                "-",
                "-f",
                f"host {ip} and tcp port {port}",
            ]


if sys.platform == "win32":

    class Dumpcap:
        @classmethod
        async def start(
            cls,
            target: TargetURI,
            artifacts_dir: Path,
        ) -> Self | None:
            logger.warn("dumpcap is not available on windows")
            return None

        async def stop(self) -> None:
            pass

        async def sync(self) -> None:
            pass



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import random
import sys
from typing import Literal

assert sys.platform.startswith("linux"), "unsupported platform"

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, AutoLiteral, Field, HexBytes, Ranges
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds.core.client import UDSRequestConfig
from gallia.services.uds.core.constants import UDSErrorCodes, UDSIsoServices
from gallia.services.uds.core.exception import IllegalResponse
from gallia.services.uds.core.service import NegativeResponse, UDSResponse
from gallia.services.uds.helpers import suggests_identifier_not_supported
from gallia.transports import RawCANTransport, TargetURI
from gallia.utils import handle_task_error, set_task_handler_ctx_variable

logger = get_logger(__name__)


class PDUFuzzerConfig(UDSScannerConfig):
    sessions: Ranges = Field([1], description="Set list of sessions to be tested; 0x01 if None")
    service: AutoLiteral[
        Literal[UDSIsoServices.WriteDataByIdentifier, UDSIsoServices.RoutineControl]
    ] = Field(
        UDSIsoServices.WriteDataByIdentifier,
        description="Service ID to create payload for; defaults to 0x2e WriteDataByIdentifier;\ncurrently supported:\n0x2e WriteDataByIdentifier, 0x31 RoutineControl (startRoutine)\n",
    )
    max_length: AutoInt = Field(42, description="maximum length of the payload")
    min_length: AutoInt = Field(1, description="minimum length of the payload")
    iterations: AutoInt = Field(1, description="number of iterations", short="i")
    dids: Ranges = Field(description="data identifiers to fuzz")
    prefixed_payload: HexBytes = Field(
        b"", description="static payload, which precedes the fuzzed payload", metavar="HEXSTRING"
    )
    observe_can_ids: Ranges = Field([], description="can ids to observe while fuzzing")


class PDUFuzzer(UDSScanner):
    """Payload fuzzer"""

    CONFIG_TYPE = PDUFuzzerConfig
    SHORT_HELP = "fuzz the UDS pdu of selected services"

    def __init__(self, config: PDUFuzzerConfig):
        super().__init__(config)
        self.config: PDUFuzzerConfig = config

    def generate_payload(self) -> bytes:
        return random.randbytes(random.randint(self.config.min_length, self.config.max_length))

    async def observe_can_messages(self, can_ids: list[int]) -> None:
        can_url = self.config.target.url._replace(scheme=RawCANTransport.SCHEME)
        transport = await RawCANTransport.connect(TargetURI(can_url.geturl()))
        transport.set_filter(can_ids, inv_filter=False)

        try:
            can_msgs: dict[int, bytes] = {}
            logger.debug("Started observe messages task")
            while True:
                try:
                    can_id, msg = await transport.recvfrom(timeout=1)
                    if can_id in can_msgs:
                        if msg != can_msgs[can_id]:
                            logger.result(f"Message for {can_id:03x} changed to {msg.hex()}")
                            can_msgs[can_id] = msg
                    else:
                        can_msgs[can_id] = msg
                        logger.result(f"Observed new message from {can_id:03x}: {msg.hex()}")
                except TimeoutError:
                    continue

        except asyncio.CancelledError:
            logger.debug("Can message observer task cancelled")

    async def main(self) -> None:
        if len(self.config.observe_can_ids) > 0:
            recv_task = asyncio.create_task(self.observe_can_messages(self.config.observe_can_ids))
            recv_task.add_done_callback(
                handle_task_error,
                context=set_task_handler_ctx_variable(__name__, "ReceiveTask"),
            )

        logger.info(f"testing sessions {self.config.sessions}")

        for did in self.config.dids:
            if self.config.service == UDSIsoServices.RoutineControl:
                pdu = bytes([self.config.service.value, 0x01, did >> 8, did & 0xFF])
            elif self.config.service == UDSIsoServices.WriteDataByIdentifier:
                pdu = bytes([self.config.service.value, did >> 8, did & 0xFF])
            for session in self.config.sessions:
                logger.notice(f"Switching to session 0x{session:02x}")
                resp: UDSResponse = await self.ecu.set_session(session)
                if isinstance(resp, NegativeResponse):
                    logger.warning(f"Switching to session 0x{session:02x} failed: {resp}")
                    continue

                logger.result(f"Starting scan in session: 0x{session:02x}")
                positive_DIDs = 0
                negative_responses: dict[UDSErrorCodes, int] = {}
                timeout_DIDs = 0
                illegal_resp = 0
                flow_control_miss = 0

                for _ in range(self.config.iterations):
                    payload = self.config.prefixed_payload + self.generate_payload()
                    try:
                        resp = await self.ecu.send_raw(
                            pdu + payload, config=UDSRequestConfig(tags=["ANALYZE"], max_retry=3)
                        )

                        if isinstance(resp, NegativeResponse):
                            if not suggests_identifier_not_supported(resp):
                                logger.result(f"0x{did:0x}: {resp}")
                            else:
                                logger.info(f"0x{did:0x}: {resp}")
                            if resp.response_code in negative_responses:
                                negative_responses[resp.response_code] += 1
                            else:
                                negative_responses[resp.response_code] = 1
                        else:
                            logger.result(f"0x{did:0x}: {resp}")
                            positive_DIDs += 1

                    except TimeoutError:
                        logger.warning(f"0x{did:0x}: Retries exceeded")
                        timeout_DIDs += 1
                    except IllegalResponse as e:
                        logger.warning(f"{repr(e)}")
                        illegal_resp += 1
                    # Temporary patch: Exception handler is deleted when it goes productive
                    except ConnectionError:
                        logger.warning("isotp flow control frame missing. Reconnecting…")
                        flow_control_miss += 1
                        await self.ecu.reconnect()

                logger.result(f"Scan in session 0x{session:0x} is complete!")

                for k, v in negative_responses.items():
                    logger.result(f"{k.name}: {v}")

                logger.result(f"Positive replies: {positive_DIDs}")

                logger.result(f"Timeouts: {timeout_DIDs}")
                logger.result(f"Illegal replies: {illegal_resp}")
                logger.result(f"Flow control frames missing: {flow_control_miss}")

                logger.info(f"Leaving session 0x{session:02x} via hook")
                await self.ecu.leave_session(session, sleep=self.config.power_cycle_sleep)

        if len(self.config.observe_can_ids) > 0:
            recv_task.cancel()
            await recv_task



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio

from gallia.command import UDSDiscoveryScanner
from gallia.command.config import AutoInt, Field
from gallia.command.uds import UDSDiscoveryScannerConfig
from gallia.log import get_logger
from gallia.services.uds.core.service import (
    DiagnosticSessionControlRequest,
    DiagnosticSessionControlResponse,
    UDSRequest,
)
from gallia.services.uds.helpers import raise_for_mismatch
from gallia.transports.base import TargetURI
from gallia.transports.hsfz import HSFZConnection

logger = get_logger(__name__)


class HSFZDiscovererConfig(UDSDiscoveryScannerConfig):
    reversed: bool = Field(False, description="scan in reversed order")
    src_addr: AutoInt = Field(0xF4, description="HSFZ source address")
    start: AutoInt = Field(0x00, description="set start address", metavar="INT")
    stop: AutoInt = Field(0xFF, description="set end address", metavar="INT")


class HSFZDiscoverer(UDSDiscoveryScanner):
    """ECU and routing discovery scanner for HSFZ"""

    SHORT_HELP = ""

    CONFIG_TYPE = HSFZDiscovererConfig

    def __init__(self, config: HSFZDiscovererConfig):
        super().__init__(config)
        self.config: HSFZDiscovererConfig = config

    async def _probe(self, conn: HSFZConnection, req: UDSRequest, timeout: float) -> bool:
        data = req.pdu
        result = False

        await asyncio.wait_for(conn.write_diag_request(data), timeout)

        # Broadcast endpoints deliver more responses.
        # Make sure to flush the receive queue properly.
        while True:
            try:
                raw_resp = await asyncio.wait_for(conn.read_diag_request(), timeout)
            except TimeoutError:
                return result

            resp = DiagnosticSessionControlResponse.parse_static(raw_resp)
            raise_for_mismatch(req, resp)
            result = True

    async def probe(
        self,
        host: str,
        port: int,
        src_addr: int,
        dst_addr: int,
        timeout: float,
        ack_timeout: float = 1.0,
    ) -> TargetURI | None:
        req = DiagnosticSessionControlRequest(0x01)

        try:
            conn = await HSFZConnection.connect(host, port, src_addr, dst_addr, ack_timeout)
        except TimeoutError:
            return None

        try:
            result = await self._probe(conn, req, timeout)
        except (TimeoutError, ConnectionError):
            return None
        finally:
            await conn.close()

        if result:
            return TargetURI.from_parts(
                "hsfz",
                host,
                port,
                {
                    "src_addr": f"{src_addr:#02x}",
                    "dst_addr": f"{dst_addr:#02x}",
                    "ack_timeout": int(ack_timeout) * 1000,
                },
            )
        return None

    async def main(self) -> None:
        found = []
        gen = (
            range(self.config.stop + 1, self.config.start)
            if self.config.reversed
            else range(self.config.start, self.config.stop + 1)
        )

        for dst_addr in gen:
            logger.info(f"testing target {dst_addr:#02x}")

            hostname = self.config.target.hostname
            port = self.config.target.port

            assert hostname is not None
            assert port is not None

            target = await self.probe(
                hostname,
                port,
                self.config.src_addr,
                dst_addr,
                self.config.timeout,
            )

            if target is not None:
                logger.info(f"found {dst_addr:#02x}")
                found.append(target)

        logger.result(f"Found {len(found)} targets")

        if self.artifacts_dir is not None:
            ecus_file = self.artifacts_dir.joinpath("ECUs.txt")
            logger.result(f"Writing urls to file: {ecus_file}")
            with ecus_file.open("w") as f:
                for target in found:
                    f.write(f"{target}\n")

        if self.db_handler is not None:
            logger.result("Writing urls to database")
            for target in found:
                await self.db_handler.insert_discovery_result(str(target))



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

from abc import ABC

from pydantic import field_validator

from gallia.command.base import FileNames, Scanner, ScannerConfig
from gallia.command.config import Field
from gallia.log import get_logger
from gallia.plugins.plugin import load_ecu, load_ecus
from gallia.services.uds.core.service import NegativeResponse, UDSResponse
from gallia.services.uds.ecu import ECU
from gallia.services.uds.helpers import raise_for_error
from gallia.transports.base import BaseTransport

logger = get_logger(__name__)


class UDSScannerConfig(ScannerConfig, cli_group="uds", config_section="gallia.protocols.uds"):
    ecu_reset: int | None = Field(
        None,
        description="Trigger an initial ecu_reset via UDS; reset level is optional",
        const=0x01,
    )
    oem: str = Field(
        ECU.OEM,
        description="The OEM of the ECU, used to choose a OEM specific ECU implementation",
        metavar="OEM",
    )
    timeout: float = Field(
        2, description="Timeout value to wait for a response from the ECU", metavar="SECONDS"
    )
    max_retries: int = Field(
        3,
        description="Number of maximum retries while sending UDS requests. If supported by the transport, this will trigger reconnects if required.",
        metavar="INT",
    )
    ping: bool = Field(True, description="Enable/Disable initial TesterPresent request")
    tester_present_interval: float = Field(
        0.5,
        description="Modify the interval of the cyclic tester present packets",
        metavar="SECONDS",
    )
    tester_present: bool = Field(
        True, description="Enable/Disable tester present background worker"
    )
    properties: bool = Field(
        True, description="Read and store the ECU proporties prior and after scan"
    )
    compare_properties: bool = Field(
        True, description="Compare properties before and after the scan"
    )

    @field_validator("oem")
    @classmethod
    def check_oem(cls, v: str) -> str:
        ecu_names = [ecu.OEM for ecu in load_ecus()]

        if v not in ecu_names:
            raise ValueError(f"Not a valid OEM. Use any of {ecu_names}.")

        return v


class UDSScanner(Scanner, ABC):
    """UDSScanner is a baseclass, particularly for scanning tasks
    related to the UDS protocol. The differences to Scanner are:

    - `self.ecu` contains a OEM specific UDS client object.
    - A background tasks sends TesterPresent regularly to avoid timeouts.
    """

    SUBGROUP: str | None = "uds"

    def __init__(self, config: UDSScannerConfig):
        super().__init__(config)
        self.config: UDSScannerConfig = config
        self._ecu: ECU | None = None
        self._implicit_logging = True

    @property
    def transport(self) -> BaseTransport:
        if self._ecu is None:
            logger.debug(
                "Transport is accessed without initialized ECU, returning Scanner transport!"
            )
            return super().transport
        return self.ecu.transport

    @transport.setter
    def transport(self, transport: BaseTransport) -> None:
        if self._ecu is None:
            logger.debug(
                "Transport is accessed without initialized ECU, setting Scanner transport!"
            )
            self._transport = transport
        else:
            self.ecu.transport = transport

    @property
    def ecu(self) -> ECU:
        if self._ecu is None:
            raise RuntimeError("ECU accessed before first initialization!")
        return self._ecu

    @ecu.setter
    def ecu(self, ecu: ECU) -> None:
        self._ecu = ecu
        # An initialized ECU has its own transport, no need to maintain a copy!
        self._transport = None

    @property
    def implicit_logging(self) -> bool:
        return self._implicit_logging

    @implicit_logging.setter
    def implicit_logging(self, value: bool) -> None:
        self._implicit_logging = value

        if self.db_handler is not None:
            self._apply_implicit_logging_setting()

    def _apply_implicit_logging_setting(self) -> None:
        self.ecu.implicit_logging = self._implicit_logging

    async def setup(self) -> None:
        await super().setup()

        self.ecu = load_ecu(self.config.oem)(
            self.transport,
            timeout=self.config.timeout,
            max_retry=self.config.max_retries,
            power_supply=self.power_supply,
        )

        self.ecu.db_handler = self.db_handler

        if self.db_handler is not None:
            try:
                # No idea, but str(args.target) fails with a strange traceback.
                # Lets use the attribute directly…
                await self.db_handler.insert_scan_run(self.config.target.raw)
                self._apply_implicit_logging_setting()
            except Exception as e:
                logger.warning(f"Could not write the scan run to the database: {e:!r}")

        if self.config.ecu_reset is not None:
            resp: UDSResponse = await self.ecu.ecu_reset(self.config.ecu_reset)
            if isinstance(resp, NegativeResponse):
                logger.warning(f"ECUReset failed: {resp}")
                logger.warning("Switching to default session")
                raise_for_error(await self.ecu.set_session(0x01))
                resp = await self.ecu.ecu_reset(self.config.ecu_reset)
                if isinstance(resp, NegativeResponse):
                    logger.warning(f"ECUReset in session 0x01 failed: {resp}")

        # Handles connecting to the target and waits
        # until it is ready.
        if self.config.ping:
            await self.ecu.wait_for_ecu()

        await self.ecu.connect()

        if self.config.tester_present:
            await self.ecu.start_cyclic_tester_present(self.config.tester_present_interval)

        if self.config.properties is True:
            properties = await self.ecu.properties(True)

            if self.artifacts_dir is not None:
                path = self.artifacts_dir.joinpath(FileNames.PROPERTIES_PRE.value)
                path.write_text(properties.to_json(indent=4) + "\n")

            if self.db_handler is not None:
                self._apply_implicit_logging_setting()

                try:
                    await self.db_handler.insert_scan_run_properties_pre(properties)
                except Exception as e:
                    logger.warning(f"Could not write the properties_pre to the database: {e!r}")

    async def teardown(self) -> None:
        if self.config.properties is True and (not self.ecu.transport.is_closed):
            properties = await self.ecu.properties(True)

            if self.artifacts_dir is not None:
                prop_curr = properties.to_json(indent=4) + "\n"
                path = self.artifacts_dir.joinpath(FileNames.PROPERTIES_POST.value)
                path.write_text(prop_curr)

                path_pre = self.artifacts_dir.joinpath(FileNames.PROPERTIES_PRE.value)
                prop_pre = path_pre.read_text()

                if self.config.compare_properties and prop_curr != prop_pre:
                    logger.warning("ecu properties differ, please investigate!")

            if self.db_handler is not None:
                try:
                    await self.db_handler.complete_scan_run(properties)
                except Exception as e:
                    logger.warning(f"Could not write the scan run to the database: {e!r}")

        if self.config.tester_present:
            await self.ecu.stop_cyclic_tester_present()

        # self.ecu.transport will be different from self.transport if self.ecu.reconnect() was called at any time
        # It is important to close this new transport as well!
        logger.debug("Closing transport object of ECU/UDSClient")
        await self.ecu.transport.close()

        # This must be the last one.
        await super().teardown()


class UDSDiscoveryScannerConfig(ScannerConfig):
    timeout: float = Field(0.5, description="timeout value for request")


class UDSDiscoveryScanner(Scanner, ABC):
    def __init__(self, config: UDSDiscoveryScannerConfig):
        super().__init__(config)
        self.config: UDSDiscoveryScannerConfig = config

    async def setup(self) -> None:
        await super().setup()

        if self.db_handler is not None:
            try:
                await self.db_handler.insert_discovery_run(self.config.target.url.scheme)
            except Exception as e:
                logger.warning(f"Could not write the discovery run to the database: {e!r}")



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds import NegativeResponse, UDSResponse
from gallia.services.uds.core.utils import g_repr

logger = get_logger(__name__)


class ECUResetPrimitiveConfig(UDSScannerConfig):
    properties: bool = Field(
        False,
        description="Read and store the ECU proporties prior and after scan",
        cli_group=UDSScannerConfig._cli_group,
        config_section=UDSScannerConfig._config_section,
    )
    session: AutoInt = Field(0x01, description="set session perform test in")
    subfunc: AutoInt = Field(0x01, description="subfunc", short="f")


class ECUResetPrimitive(UDSScanner):
    """Use the ECUReset UDS service to reset the ECU"""

    CONFIG_TYPE = ECUResetPrimitiveConfig
    SHORT_HELP = "ECUReset"

    def __init__(self, config: ECUResetPrimitiveConfig):
        super().__init__(config)
        self.config: ECUResetPrimitiveConfig = config

    async def main(self) -> None:
        resp: UDSResponse = await self.ecu.set_session(self.config.session)
        if isinstance(resp, NegativeResponse):
            logger.error(f"could not change to session: {g_repr(self.config.session)}")
            return

        try:
            logger.info(f"try sub-func: {g_repr(self.config.subfunc)}")
            resp = await self.ecu.ecu_reset(self.config.subfunc)
            if isinstance(resp, NegativeResponse):
                msg = f"ECU Reset {g_repr(self.config.subfunc)} failed in session: {g_repr(self.config.session)}: {resp}"
                logger.error(msg)
            else:
                logger.result(f"ECU Reset {g_repr(self.config.subfunc)} succeeded")
        except TimeoutError:
            logger.error("Timeout")
            await asyncio.sleep(10)
        except ConnectionError:
            msg = f"Lost connection to ECU, session: {g_repr(self.config.session)} subFunc: {g_repr(self.config.subfunc)}"
            logger.error(msg)
            return



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import importlib
import json
import sys
from collections.abc import Mapping
from pathlib import Path
from typing import Any, Self

import aiosqlite
from pydantic import model_validator

from gallia.command import BaseCommand
from gallia.command.base import AsyncScript, AsyncScriptConfig
from gallia.command.config import Field
from gallia.log import get_logger

logger = get_logger(__name__)


class RerunnerConfig(AsyncScriptConfig):
    id: int | None = Field(None, description="The id of the run_meta entry in the db")
    file: Path | None = Field(None, description="The path of the META.json in the logs")

    @model_validator(mode="after")
    def check_transport_requirements(self) -> Self:
        if self.id is not None and self.db is None:
            raise ValueError("This script requires a database connection")

        return self

    @model_validator(mode="after")
    def check_meta_source(self) -> Self:
        if not (self.id is None) ^ (self.file is None):
            raise ValueError("Exactly one of id or file is required")

        return self


class Rerunner(AsyncScript):
    CONFIG_TYPE = RerunnerConfig
    SHORT_HELP = "Rerun a previous gallia command based on its run_meta in the database"

    def __init__(self, config: RerunnerConfig):
        super().__init__(config)
        self.config: RerunnerConfig = config

    async def main(self) -> None:
        if self.config.id is not None:
            script, config = await self.db()
        else:
            script, config = self.file()

        # TODO: Make a interface for this to avoid error-prone parsing.
        script_parts = script.split(".")
        module = ".".join(script_parts[:-1])
        class_name = script_parts[-1]

        logger.info(f"Rerunning run {self.config.id} ({class_name}) with: {config}")

        gallia_class: type[BaseCommand] = getattr(importlib.import_module(module), class_name)
        command = gallia_class(gallia_class.CONFIG_TYPE(**config))

        sys.exit(await command.entry_point())

    async def db(self) -> tuple[str, Mapping[str, Any]]:
        assert self.config.id is not None

        query = "SELECT script, config FROM run_meta WHERE id = ?"
        parameters = (self.config.id,)

        assert self.db_handler is not None

        connection = self.db_handler.connection

        assert connection is not None

        cursor: aiosqlite.Cursor = await connection.execute(query, parameters)
        row = await cursor.fetchone()

        if row is None:
            logger.error(f"There id no run_meta entry with the id {self.config.id}")
            sys.exit(1)

        return row[0], json.loads(row[1])

    def file(self) -> tuple[str, Mapping[str, Any]]:
        assert self.config.file is not None

        with self.config.file.open("r") as f:
            content = json.load(f)

        return content["command"], content["config"]



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

from functools import partial

from gallia.log import get_logger
from gallia.transports import TargetURI

logger = get_logger(__name__)


class PowerSupplyURI(TargetURI):
    @property
    def id(self) -> int:
        if "id" in self.qs:
            return int(self.qs["id"][0], 0)
        return 0

    @property
    def channel(self) -> int | list[int]:
        if "channel" in self.qs:
            if len(ch := self.qs["channel"]) == 1:
                return int(ch[0], 0)
            return list(map(partial(int, base=0), ch))
        return 0

    @property
    def product_id(self) -> str:
        if "product_id" in self.qs:
            return self.qs["product_id"][0].lower()
        return ""



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import errno
import struct
import sys
from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Self

from pydantic import BaseModel, field_validator

from gallia.log import get_logger
from gallia.transports.base import BaseTransport, TargetURI
from gallia.utils import auto_int, handle_task_error, set_task_handler_ctx_variable

logger = get_logger(__name__)


class HSFZStatus(IntEnum):
    UNDEFINED = -0x01
    Data = 0x01
    Ack = 0x02
    Klemme15 = 0x10
    Vin = 0x11
    AliveCheck = 0x12
    StatusDataInquiry = 0x13
    IncorrectTesterAddressError = 0x40
    IncorrectControlWordError = 0x41
    IncorrectFormatError = 0x42
    IncorrectDestinationAddressError = 0x43
    MessageTooLarge = 0x44
    ApplicationNotReady = 0x45
    OutOfMemory = 0xFF

    @classmethod
    def _missing_(cls, value: Any) -> "HSFZStatus":
        return cls.UNDEFINED


@dataclass
class HSFZHeader:
    Len: int
    CWord: int

    def pack(self) -> bytes:
        return struct.pack("!IH", self.Len, self.CWord)

    @classmethod
    def unpack(cls, data: bytes) -> Self:
        len_, cword = struct.unpack("!IH", data)
        return cls(len_, cword)


@dataclass
class HSFZDiagReqHeader:
    src_addr: int
    dst_addr: int

    def pack(self) -> bytes:
        return struct.pack("!BB", self.src_addr, self.dst_addr)

    @classmethod
    def unpack(cls, data: bytes) -> Self:
        src_addr, dst_addr = struct.unpack("!BB", data)
        return cls(src_addr, dst_addr)


HSFZFrame = tuple[HSFZHeader, HSFZDiagReqHeader | None, bytes | None]
HSFZDiagFrame = tuple[HSFZHeader, HSFZDiagReqHeader, bytes]


class HSFZConnection:
    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        src_addr: int,
        dst_addr: int,
        ack_timeout: float = 1.0,
    ):
        self.reader = reader
        self.writer = writer
        self.src_addr = src_addr
        self.dst_addr = dst_addr
        self.ack_timeout = ack_timeout
        self._read_queue: asyncio.Queue[HSFZDiagFrame | int] = asyncio.Queue()
        self._read_task = asyncio.create_task(self._read_worker())
        self._read_task.add_done_callback(
            handle_task_error,
            context=set_task_handler_ctx_variable(__name__, "HsfzReader"),
        )
        self._closed = False
        self._mutex = asyncio.Lock()

    @classmethod
    async def connect(
        cls,
        host: str,
        port: int,
        src_addr: int,
        dst_addr: int,
        ack_timeout: float,
    ) -> Self:
        reader, writer = await asyncio.open_connection(host, port)
        return cls(
            reader,
            writer,
            src_addr,
            dst_addr,
            ack_timeout,
        )

    async def _read_frame(self) -> HSFZFrame:
        # Header is fixed size 6 byte.
        hdr_buf = await self.reader.readexactly(6)
        hdr = HSFZHeader.unpack(hdr_buf)

        # If a message without a RequestHeader is received,
        # the whole message must be read before erroring out.
        # Otherwise the partial read packet stays in the receive
        # buffer and causes further breakage…
        if hdr.Len < 2:
            data = None
            if hdr.Len > 0:
                data = await self.reader.readexactly(hdr.Len)
            data_str = data.hex() if data is not None else data
            logger.trace(f"hdr: {hdr}, req_hdr: None, data: {data_str}", extra={"tags": ["read"]})
            return hdr, None, data

        # DiagReqHeader is fixed size 2 byte.
        req_buf = await self.reader.readexactly(2)
        req_hdr = HSFZDiagReqHeader.unpack(req_buf)

        data_len = hdr.Len - 2
        data = await self.reader.readexactly(data_len)
        logger.trace(
            f"hdr: {hdr}, req_hdr: {req_hdr}, data: {data.hex()}",
            extra={"tags": ["read"]},
        )
        return hdr, req_hdr, data

    async def write_frame(self, frame: HSFZFrame) -> None:
        hdr, req_hdr, data = frame
        buf = b""
        buf += hdr.pack()
        log_msg = f"hdr: {hdr}"
        if req_hdr is not None:
            buf += req_hdr.pack()
            log_msg += f", req_hdr: {req_hdr}"
            if data is not None:
                buf += data
                log_msg += f", data: {data.hex()}"
        self.writer.write(buf)
        await self.writer.drain()

        logger.trace(log_msg, extra={"tags": ["write"]})

    async def _read_worker(self) -> None:
        try:
            while True:
                hdr, req_hdr, data = await self._read_frame()

                match hdr.CWord:
                    case HSFZStatus.AliveCheck:
                        await self.send_alive_msg()
                        continue
                    case HSFZStatus.Ack | HSFZStatus.Data:
                        if req_hdr is None:
                            logger.warning("unexpected frame: no hsfz request header")
                            continue
                        if data is None:
                            logger.warning("unexpected frame: no payload")
                            continue
                        await self._read_queue.put((hdr, req_hdr, data))
                    case _:
                        await self._read_queue.put(hdr.CWord)
                        continue

        except asyncio.CancelledError:
            logger.debug("read worker cancelled")
        except asyncio.IncompleteReadError as e:
            logger.debug(f"read worker received EOF: {e}")
        except Exception as e:
            logger.critical(f"read worker died: {e}")

    async def _unpack_frame(self, frame: HSFZDiagFrame | int) -> HSFZDiagFrame:
        # I little hack, but it is either a tuple or an int….
        match frame:
            case tuple():
                return frame
            case int():
                await self.close()
                raise BrokenPipeError(f"I can't even: {HSFZStatus(frame).name}")
            case _:
                raise RuntimeError(f"unexpected frame: {frame}")

    async def read_frame(self) -> HSFZDiagFrame | int:
        if self._closed:
            if sys.platform != "win32":
                raise OSError(errno.EBADFD)
            else:
                raise RuntimeError("connection already closed")

        return await self._read_queue.get()

    async def read_diag_request(self) -> bytes:
        unexpected_packets = []
        while True:
            hdr, req_hdr, data = await self._unpack_frame(await self.read_frame())
            if hdr.CWord != HSFZStatus.Data:
                logger.warning(
                    f"expected HSFZ data, instead got: {HSFZStatus(hdr.CWord).name} with payload {data.hex()}"
                )
                unexpected_packets.append((hdr, req_hdr, data))
                continue
            if req_hdr.src_addr != self.dst_addr or req_hdr.dst_addr != self.src_addr:
                logger.warning(
                    f"HSFZ Data has unexpected addresses (src:dst); should be {self.dst_addr:#04x}:{self.src_addr:#04x}, but is {req_hdr.src_addr:#04x}:{req_hdr.dst_addr:#04x}"
                )
                unexpected_packets.append((hdr, req_hdr, data))
                continue

            # We do not want to consume packets that we were not expecting; add them to queue again
            for item in unexpected_packets:
                await self._read_queue.put(item)

            return data

    async def _read_ack(self, prev_data: bytes) -> None:
        unexpected_packets = []
        while True:
            hdr, req_hdr, data = await self._unpack_frame(await self.read_frame())
            if hdr.CWord != HSFZStatus.Ack:
                logger.warning(
                    f"expected HSFZ Ack for {prev_data.hex()}, instead got: {HSFZStatus(hdr.CWord).name} with payload {data.hex()}"
                )
                unexpected_packets.append((hdr, req_hdr, data))
                continue
            if req_hdr.src_addr != self.src_addr or req_hdr.dst_addr != self.dst_addr:
                logger.warning(
                    f"HSFZ Ack has unexpected addresses (src:dst); should be {self.src_addr:#04x}:{self.dst_addr:#04x}, but is {req_hdr.src_addr:#04x}:{req_hdr.dst_addr:#04x}"
                )
                unexpected_packets.append((hdr, req_hdr, data))
                continue
            if prev_data[:5] != data:
                logger.warning(
                    f"HSFZ Ack has unexpected data of {data.hex()}, should be {prev_data[:5].hex()}"
                )
                unexpected_packets.append((hdr, req_hdr, data))
                continue

            # We do not want to consume packets that we were not expecting; add them to queue again
            for item in unexpected_packets:
                await self._read_queue.put(item)

            return

    async def write_diag_request_raw(
        self,
        hdr: HSFZHeader,
        req_hdr: HSFZDiagReqHeader,
        data: bytes,
    ) -> None:
        async with self._mutex:
            await self.write_frame((hdr, req_hdr, data))

            try:
                # Now an ACK message is expected.
                await asyncio.wait_for(self._read_ack(data), self.ack_timeout)
            except TimeoutError as e:
                await self.close()
                raise BrokenPipeError("no ack by gateway") from e

    async def write_diag_request(self, data: bytes) -> None:
        hdr = HSFZHeader(Len=len(data) + 2, CWord=HSFZStatus.Data)
        req_hdr = HSFZDiagReqHeader(src_addr=self.src_addr, dst_addr=self.dst_addr)
        await self.write_diag_request_raw(hdr, req_hdr, data)

    async def send_alive_msg(self) -> None:
        hdr = HSFZHeader(Len=2, CWord=HSFZStatus.AliveCheck)
        buf = b""
        buf += hdr.pack()
        # For reasons, the tester address is two bytes large in this path.
        buf += struct.pack("!H", self.src_addr)

        self.writer.write(buf)
        await self.writer.drain()

    async def close(self) -> None:
        if self._closed:
            return

        self._closed = True
        self._read_task.cancel()
        self.writer.close()
        await self.writer.wait_closed()


class HSFZConfig(BaseModel):
    src_addr: int
    dst_addr: int
    ack_timeout: int = 1000

    @field_validator(
        "src_addr",
        "dst_addr",
        mode="before",
    )
    def auto_int(cls, v: str) -> int:
        return auto_int(v)


class HSFZTransport(BaseTransport, scheme="hsfz"):
    def __init__(
        self,
        target: TargetURI,
        port: int,
        config: HSFZConfig,
        conn: HSFZConnection,
    ):
        super().__init__(target)
        self._conn = conn
        self.port = port

    @classmethod
    async def connect(
        cls,
        target: str | TargetURI,
        timeout: float | None = None,
    ) -> Self:
        t = TargetURI(target) if isinstance(target, str) else target
        if t.hostname is None:
            raise ValueError("no hostname specified")

        port = t.port if t.port is not None else 6801
        config = HSFZConfig.model_validate(t.qs_flat)
        conn = await HSFZConnection.connect(
            t.hostname,
            port,
            config.src_addr,
            config.dst_addr,
            config.ack_timeout / 1000,
        )
        return cls(
            t,
            port,
            config,
            conn,
        )

    async def close(self) -> None:
        await self._conn.close()

    async def read(
        self,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> bytes:
        return await asyncio.wait_for(self._conn.read_diag_request(), timeout)

    async def write(
        self,
        data: bytes,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> int:
        await asyncio.wait_for(self._conn.write_diag_request(data), timeout)
        return len(data)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

from typing import Self

from gallia.transports import TargetURI
from gallia.transports.base import BaseTransport


class DummyTransport(BaseTransport, scheme="dummy"):
    @classmethod
    async def connect(cls, target: str | TargetURI, timeout: float | None = None) -> Self:
        t = target if isinstance(target, TargetURI) else TargetURI(target)
        return cls(t)

    async def close(self) -> None:
        pass

    async def read(self, timeout: float | None = None, tags: list[str] | None = None) -> bytes:
        return b""

    async def write(
        self, data: bytes, timeout: float | None = None, tags: list[str] | None = None
    ) -> int:
        return len(data)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import socket
import struct
import sys
from abc import ABC

assert sys.platform.startswith("linux"), "unsupported platform"

from gallia.command import AsyncScript
from gallia.command.base import AsyncScriptConfig
from gallia.command.config import AutoInt, Field, Ranges
from gallia.log import get_logger
from gallia.services.uds.core.utils import bytes_repr, g_repr
from gallia.transports import RawCANTransport, TargetURI
from gallia.utils import can_id_repr

logger = get_logger(__name__)


class FindXCPConfig(AsyncScriptConfig):
    pass


class CanFindXCPConfig(FindXCPConfig):
    xcp_can_iface: str = Field(description="CAN interface used for XCP communication")
    can_fd: bool = Field(False, description="use can FD")
    extended: bool = Field(False, description="use extended CAN address space")
    sniff_time: int = Field(
        60, description="Time in seconds to sniff on bus for current traffic", metavar="SECONDS"
    )
    can_id_start: AutoInt = Field(description="First CAN id to test")
    can_id_end: AutoInt = Field(description="Last CAN id to test")


class TcpFindXCPConfig(FindXCPConfig):
    xcp_ip: str = Field(description="XCP destination IP Address")
    tcp_ports: Ranges = Field(description="List of TCP ports to test for XCP")


class UdpFindXCPConfig(FindXCPConfig):
    xcp_ip: str = Field(description="XCP destination IP Address")
    udp_ports: Ranges = Field(description="List of UDP ports to test for XCP")


class FindXCP(AsyncScript, ABC):
    """Find XCP Slave"""

    def __init__(self, config: FindXCPConfig):
        super().__init__(config)
        self.config: FindXCPConfig = config
        self.socket: socket.socket

    def pack_xcp_eth(self, data: bytes, ctr: int = 0) -> bytes:
        length = len(data)
        data = struct.pack("<HH", length, ctr) + data
        logger.info(f"send: {data.hex()}")
        return data

    def unpack_xcp_eth(self, data: bytes) -> tuple[int, int, bytes]:
        length, ctr = struct.unpack_from("<HH", data)
        logger.info(f"recv: {data.hex()}")
        return (length, ctr, data[4:])

    def xcp_disconnect(self, server: tuple[str, int]) -> None:
        try:
            self.socket.sendto(self.pack_xcp_eth(bytes([0xFE, 0x00]), 1), server)
            self.socket.recv(1024)
        except Exception:
            pass


class CanFindXCP(FindXCP):
    CONFIG_TYPE = CanFindXCPConfig
    SHORT_HELP = "XCP enumeration scanner for CAN"

    def __init__(self, config: CanFindXCPConfig):
        super().__init__(config)
        self.config: CanFindXCPConfig = config

    async def main(self) -> None:
        target = TargetURI(
            f"{RawCANTransport.SCHEME}://{self.config.xcp_can_iface}?is_extended={str(self.config.extended).lower()}"
            + ("&is_fd=true" if self.config.can_fd else "")
        )
        transport = await RawCANTransport.connect(target)
        endpoints = []

        sniff_time: int = self.config.sniff_time
        logger.result(f"Listening to idle bus communication for {sniff_time}s...")
        addr_idle = await transport.get_idle_traffic(sniff_time)
        logger.result(f"Found {len(addr_idle)} CAN Addresses on idle Bus")
        transport.set_filter(addr_idle, inv_filter=True)
        # flush receive queue
        await transport.get_idle_traffic(2)

        for can_id in range(self.config.can_id_start, self.config.can_id_end + 1):
            logger.info(f"Testing CAN ID: {can_id_repr(can_id)}")
            pdu = bytes([0xFF, 0x00])
            await transport.sendto(pdu, can_id, timeout=0.1)

            try:
                while True:
                    master, data = await transport.recvfrom(timeout=0.1)
                    if data[0] == 0xFF:
                        msg = f"Found XCP endpoint [master:slave]: CAN: {can_id_repr(master)}:{can_id_repr(can_id)} data: {bytes_repr(data)}"
                        logger.result(msg)
                        endpoints.append((can_id, master))
                    else:
                        logger.info(
                            f"Received non XCP answer for CAN-ID {can_id_repr(can_id)}: {can_id_repr(master)}:{bytes_repr(data)}"
                        )
            except TimeoutError:
                pass

        logger.result(f"Finished; Found {len(endpoints)} XCP endpoints via CAN")


class TcpFindXCP(FindXCP):
    CONFIG_TYPE = TcpFindXCPConfig
    SHORT_HELP = "XCP enumeration scanner for TCP"

    def __init__(self, config: TcpFindXCPConfig):
        super().__init__(config)
        self.config: TcpFindXCPConfig = config

    async def main(self) -> None:
        # TODO: rewrite as async

        data = bytes([0xFF, 0x00])
        endpoints = []
        for port in self.config.tcp_ports:
            logger.info(f"Testing TCP port: {port}")
            server = (self.config.xcp_ip, port)
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.settimeout(0.2)
            try:
                self.socket.connect(server)
            except Exception as e:
                logger.info(f"Connect: {g_repr(e)} on TCP port {port}")
                continue

            try:
                self.socket.send(self.pack_xcp_eth(data))
                _, _, data_ret = self.unpack_xcp_eth(self.socket.recv(1024))
                ret = bytes_repr(data_ret)
                logger.info(f"Receive data on TCP port {port}: {ret}")
                if len(data_ret) > 0 and data_ret[0] == 0xFF:
                    logger.result(f"XCP Slave on TCP port {port}, data: {ret}")
                    endpoints.append(port)
                else:
                    logger.info(f"TCP port {port} is no XCP slave, data: {ret}")
            except Exception as e:
                logger.info(f"send/recv: {g_repr(e)} on TCP port {port:d}")
                continue
            finally:
                self.xcp_disconnect(server)
                self.socket.close()

        logger.result(f"Finished; Found {len(endpoints)} XCP endpoints via TCP")


class UdpFindXCP(FindXCP):
    CONFIG_TYPE = UdpFindXCPConfig
    SHORT_HELP = "XCP enumeration scanner for Udp"

    def __init__(self, config: UdpFindXCPConfig):
        super().__init__(config)
        self.config: UdpFindXCPConfig = config

    async def main(self) -> None:
        self.test_eth_broadcast()
        await self.test_udp()

    async def test_udp(self) -> None:
        # TODO: rewrite as async

        data = bytes([0xFF, 0x00])
        endpoints = []
        for port in self.config.udp_ports:
            logger.info(f"Testing UDP port: {port}")
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.socket.settimeout(0.5)
            server = (self.config.xcp_ip, port)
            self.socket.sendto(self.pack_xcp_eth(data), server)
            try:
                _, _, data_ret = self.unpack_xcp_eth(self.socket.recv(1024))
                ret = bytes_repr(data_ret)
                logger.info(f"Receive data on UDP port {port}: {ret}")
                if len(data_ret) > 0 and data_ret[0] == 0xFF:
                    logger.result(f"XCP Slave on UDP port {port}, data: {ret}")
                    endpoints.append(port)
                else:
                    logger.info(f"UDP port {port} is no XCP slave, data: {ret}")

            except TimeoutError:
                logger.info(f"Timeout on UDP port {port}")

            finally:
                self.xcp_disconnect(server)
                self.socket.close()

        logger.result(f"Finished; Found {len(endpoints)} XCP endpoints via UDP")

    def test_eth_broadcast(self) -> None:
        # TODO: rewrite as async

        multicast_group = ("239.255.0.0", 5556)
        logger.result(
            f"Discover XCP via multicast group: {multicast_group[0]}:{multicast_group[1]}"
        )

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.connect((self.config.xcp_ip, 5555))
        addr = self.socket.getsockname()[0]
        self.socket.close()
        logger.info(f"xcp interface ip for multicast group: {addr}")

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton(addr))
        self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 32)
        self.socket.settimeout(2)

        xcp_discover = bytes([0xFA, 0x01])
        endpoints = []
        self.socket.sendto(self.pack_xcp_eth(xcp_discover), multicast_group)
        try:
            while True:
                data, slave = self.socket.recvfrom(16)
                if not data:
                    break

                logger.result(f"Found XCP slave: {slave} {bytes_repr(data)}")
                endpoints.append(slave)
        except TimeoutError:
            logger.info("Timeout")

        logger.result(f"Finished; Found {len(endpoints)} XCP endpoints via multicast group")



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import sys

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds.core.service import NegativeResponse

logger = get_logger(__name__)


class PingPrimitiveConfig(UDSScannerConfig):
    properties: bool = Field(
        False,
        description="Read and store the ECU proporties prior and after scan",
        cli_group=UDSScannerConfig._cli_group,
        config_section=UDSScannerConfig._config_section,
    )
    session: AutoInt | None = Field(None, description="set session to perform test")
    count: AutoInt | None = Field(None, description="limit number of pings to this amount")
    interval: float = Field(0.5, description="time interval between two pings", metavar="SECONDS")


class PingPrimitive(UDSScanner):
    """Ping ECU via TesterPresent"""

    CONFIG_TYPE = PingPrimitiveConfig
    SHORT_HELP = "ping ECU via TesterPresent"

    def __init__(self, config: PingPrimitiveConfig):
        super().__init__(config)
        self.config: PingPrimitiveConfig = config

    async def main(self) -> None:
        if self.config.session is not None:
            resp = await self.ecu.set_session(self.config.session)
            if isinstance(resp, NegativeResponse):
                logger.error(f"Could not change to requested session: {resp}")
                sys.exit(1)

        i = 1
        while True:
            ret = await self.ecu.ping()
            if isinstance(ret, NegativeResponse):
                logger.warning(ret)
            logger.result("ECU is alive!")
            if self.config.count is not None and i >= self.config.count:
                break
            await asyncio.sleep(self.config.interval)
            i += 1



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys
from typing import Annotated, cast

from pydantic import BeforeValidator

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field, err_int
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds import NegativeResponse
from gallia.services.uds.core.utils import g_repr

logger = get_logger(__name__)


class DDDIPrimitiveConfig(UDSScannerConfig):
    properties: bool = Field(
        False,
        description="Read and store the ECU properties prior and after scan",
        cli_group=UDSScannerConfig._cli_group,
        config_section=UDSScannerConfig._config_section,
    )
    session: AutoInt = Field(0x01, description="The session in which the requests are made")


def parse_definitions(value: str | tuple[int, ...], expected_len: int) -> tuple[int, ...]:
    if isinstance(value, tuple):
        if len(value) != expected_len:
            raise ValueError(f"Need exactly {expected_len} values for each definition")

        return value

    values = value.split(":")

    if len(values) != expected_len:
        raise ValueError(f"Need exactly {expected_len} values for each definition")

    return tuple(err_int(x, 0) for x in values)


def parse_id(value: str | tuple[int, int, int]) -> tuple[int, int, int]:
    return cast(tuple[int, int, int], parse_definitions(value, 3))


def parse_mem(value: str | tuple[int, int]) -> tuple[int, int]:
    return cast(tuple[int, int], parse_definitions(value, 2))


class DefineByIdentifierDDDIPrimitiveConfig(DDDIPrimitiveConfig):
    data_identifier: AutoInt = Field(
        description="The new dynamically defined data identifier", positional=True
    )
    sources: list[Annotated[tuple[int, int, int], BeforeValidator(parse_id)]] = Field(
        description="The definitions of the source data to be included in the dynamically defined data. Each of them consists of a data identifier, the start byte of the corresponding data record (1-indexed) and its length in bytes",
        metavar="ID:START:LENGTH",
    )


class DefineByMemoryAddressDDDIPrimitiveConfig(DDDIPrimitiveConfig):
    data_identifier: AutoInt = Field(
        description="The new dynamically defined data identifier", positional=True
    )
    sources: list[Annotated[tuple[int, int], BeforeValidator(parse_mem)]] = Field(
        description="The definitions of the source data to be included in the dynamically defined data. each of them consists of a memory address and the length in bytes",
        metavar="ADDRESS:LENGTH",
    )
    address_format: AutoInt | None = Field(
        None,
        description="The addressAndLengthFormatIdentifier, which can be set manually or deduced automatically if not given explicitly",
    )


class ClearDynamicallyDefinedDataIdentifierDDDIPrimitiveConfig(DDDIPrimitiveConfig):
    data_identifier: AutoInt | None = Field(
        description="The dynamically defined data identifier to be cleared. Omit if all dynamically defined identifiers should be cleared."
    )


class DDDIPrimitive(UDSScanner):
    """dynamically define data identifiers"""

    CONFIG_TYPE = DDDIPrimitiveConfig
    SHORT_HELP = "DynamicallyDefineDataIdentifiers"

    def __init__(self, config: DDDIPrimitiveConfig):
        super().__init__(config)
        self.config: DDDIPrimitiveConfig = config

    async def main(self) -> None:
        try:
            await self.ecu.check_and_set_session(self.config.session)
        except Exception as e:
            logger.critical(f"Could not change to session: {g_repr(self.config.session)}: {e!r}")
            sys.exit(1)


class DefineByIdentifierDDDIPrimitive(DDDIPrimitive):
    CONFIG_TYPE = DefineByIdentifierDDDIPrimitiveConfig
    SHORT_HELP = "DefineByIdentifier"

    def __init__(self, config: DefineByIdentifierDDDIPrimitiveConfig):
        super().__init__(config)
        self.config: DefineByIdentifierDDDIPrimitiveConfig = config

    async def main(self) -> None:
        await super().main()

        source_identifiers = []
        start_positions = []
        lengths = []

        for identifier, start, length in self.config.sources:
            source_identifiers.append(identifier)
            start_positions.append(start)
            lengths.append(length)

        response = await self.ecu.define_by_identifier(
            self.config.data_identifier, source_identifiers, start_positions, lengths
        )

        if isinstance(response, NegativeResponse):
            logger.error(response)
        else:
            # There is not real data returned, only echoes
            logger.result("Success")


class DefineByMemoryAddressDDDIPrimitive(DDDIPrimitive):
    CONFIG_TYPE = DefineByMemoryAddressDDDIPrimitiveConfig
    SHORT_HELP = "DefineByMemoryAddress"

    def __init__(self, config: DefineByMemoryAddressDDDIPrimitiveConfig):
        super().__init__(config)
        self.config: DefineByMemoryAddressDDDIPrimitiveConfig = config

    async def main(self) -> None:
        await super().main()

        addresses = []
        lengths = []

        for address, length in self.config.sources:
            addresses.append(address)
            lengths.append(length)

        response = await self.ecu.define_by_memory_address(
            self.config.data_identifier, addresses, lengths, self.config.address_format
        )

        if isinstance(response, NegativeResponse):
            logger.error(response)
        else:
            # There is not real data returned, only echoes
            logger.result("Success")


class ClearDynamicallyDefinedDataIdentifierDDDIPrimitive(DDDIPrimitive):
    CONFIG_TYPE = ClearDynamicallyDefinedDataIdentifierDDDIPrimitiveConfig
    SHORT_HELP = "ClearDynamicallyDefinedDataIdentifier"

    def __init__(self, config: ClearDynamicallyDefinedDataIdentifierDDDIPrimitiveConfig):
        super().__init__(config)
        self.config: ClearDynamicallyDefinedDataIdentifierDDDIPrimitiveConfig = config

    async def main(self) -> None:
        await super().main()

        response = await self.ecu.clear_dynamically_defined_data_identifier(
            self.config.data_identifier
        )

        if isinstance(response, NegativeResponse):
            logger.error(response)
        else:
            # There is not real data returned, only echoes
            logger.result("Success")



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Parses Boolean Pydantic Fields to Command-Line Arguments.

The `boolean` module contains the `should_parse` function, which checks whether
this module should be used to parse the field, as well as the `parse_field`
function, which parses boolean `pydantic` model fields to `ArgumentParser`
command-line arguments.
"""

from typing import Any

from gallia.pydantic_argparse.argparse import actions
from gallia.pydantic_argparse.utils.pydantic import PydanticField

from .utils import SupportsAddArgument


def should_parse(field: PydanticField) -> bool:
    """Checks whether the field should be parsed as a `boolean`.

    Args:
        field (PydanticField): Field to check.

    Returns:
        bool: Whether the field should be parsed as a `boolean`.
    """
    # Check and Return
    return field.is_a(bool)


def parse_field(
    parser: SupportsAddArgument,
    field: PydanticField,
) -> None:
    """Adds boolean pydantic field to argument parser.

    Args:
        parser (argparse.ArgumentParser): Argument parser to add to.
        field (PydanticField): Field to be added to parser.
    """
    # Determine Argument Properties
    action = actions.BooleanOptionalAction

    args: dict[str, Any] = {}
    args.update(field.arg_required())
    args.update(field.arg_default())
    args.update(field.arg_const())
    args.update(field.arg_dest())

    # Add Boolean Field
    parser.add_argument(*field.arg_names(), action=action, help=field.description(), **args)



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Parses Literal Pydantic Fields to Command-Line Arguments.

The `literal` module contains the `should_parse` function, which checks whether
this module should be used to parse the field, as well as the `parse_field`
function, which parses literal `pydantic` model fields to `ArgumentParser`
command-line arguments.
"""

import argparse
from typing import Any, Literal, get_args

from gallia.pydantic_argparse.utils.field import ArgFieldInfo
from gallia.pydantic_argparse.utils.pydantic import PydanticField

from .utils import SupportsAddArgument


def should_parse(field: PydanticField) -> bool:
    """Checks whether the field should be parsed as a `literal`.

    Args:
        field (PydanticField): Field to check.

    Returns:
        bool: Whether the field should be parsed as a `literal`.
    """
    # Check and Return
    return field.is_a(Literal)


def parse_field(
    parser: SupportsAddArgument,
    field: PydanticField,
) -> None:
    """Adds enum pydantic field to argument parser.

    Args:
        parser (argparse.ArgumentParser): Argument parser to add to.
        field (PydanticField): Field to be added to parser.
    """
    # Extract Choices
    choices = get_args(field.info.annotation)

    metavar = f"{{{', '.join(str(c) for c in choices)}}}"

    if isinstance(field.info, ArgFieldInfo) and field.info.metavar is not None:
        metavar = field.info.metavar

    action = argparse._StoreAction

    args: dict[str, Any] = {}
    args.update(field.arg_required())
    args.update(field.arg_default())
    args.update(field.arg_const())
    args.update(field.arg_dest())

    # Add Literal Field
    parser.add_argument(
        *field.arg_names(), action=action, help=field.description(), metavar=metavar, **args
    )



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Argument parser utilities."""

from argparse import Action, FileType
from collections.abc import Callable, Iterable
from typing import Any, Protocol, TypeVar

T = TypeVar("T")


class SupportsAddArgument(Protocol):
    """ArgumentParser protocol that captures the base parser and argument groups."""

    def add_argument(  # noqa: D102
        self,
        *name_or_flags: str,
        action: str | type[Action] = ...,
        nargs: int | str = ...,
        const: Any = ...,
        default: Any = ...,
        type: Callable[[str], T] | FileType = ...,  # noqa: A002
        choices: Iterable[T] | None = ...,
        required: bool = ...,
        help: str | None = ...,  # noqa: A002
        metavar: str | tuple[str, ...] | None = ...,
        dest: str | None = ...,
        version: str = ...,
        **kwargs: Any,
    ) -> Action: ...



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys

from tabulate import tabulate

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field, HexInt
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds.core.constants import (
    CDTCSSubFuncs,
    DiagnosticSessionControlSubFuncs,
    UDSErrorCodes,
)
from gallia.services.uds.core.service import NegativeResponse
from gallia.services.uds.core.utils import g_repr

logger = get_logger(__name__)


class DTCPrimitiveConfig(UDSScannerConfig):
    properties: bool = Field(
        False,
        description="Read and store the ECU proporties prior and after scan",
        cli_group=UDSScannerConfig._cli_group,
        config_section=UDSScannerConfig._config_section,
    )
    session: AutoInt = Field(
        DiagnosticSessionControlSubFuncs.defaultSession.value,
        description="Session to perform test in",
    )


class ReadDTCPrimitiveConfig(DTCPrimitiveConfig):
    mask: HexInt = Field(
        0xFF,
        description="The bitmask which is sent to the ECU in order to select the relevant DTCs according to their error state. By default, all error codes are returned (c.f. ISO 14229-1,D.2).",
    )
    show_legend: bool = Field(
        False, description="Show the legend of the bit interpretation according to ISO 14229-1,D.2"
    )
    show_failed: bool = Field(False, description="Show a summary of the codes which failed")
    show_uncompleted: bool = Field(
        False, description="Show a summary of the codes which have not completed"
    )


class ClearDTCPrimitiveConfig(DTCPrimitiveConfig):
    group_of_dtc: int = Field(
        0xFFFFFF,
        description="Only clear a particular DTC or the DTCs belonging to the given group. By default, all error codes are cleared.",
    )


class ControlDTCPrimitiveConfig(DTCPrimitiveConfig):
    stop: bool = Field(
        False, description="Stop the setting of DTCs. If already disabled, this has no effect."
    )
    resume: bool = Field(
        False, description="Resume the setting of DTCs. If already enabled, this has no effect."
    )


class ReadDTCPrimitive(UDSScanner):
    """Read out the Diagnostic Trouble Codes (DTC)"""

    CONFIG_TYPE = ReadDTCPrimitiveConfig
    SHORT_HELP = "Read the DTCs using the ReadDTCInformation service"

    def __init__(self, config: ReadDTCPrimitiveConfig):
        super().__init__(config)
        self.config: ReadDTCPrimitiveConfig = config

    async def fetch_error_codes(self, mask: int, split: bool = True) -> dict[int, int]:
        ecu_response = await self.ecu.read_dtc_information_report_dtc_by_status_mask(mask)
        dtcs = {}

        if isinstance(ecu_response, NegativeResponse):
            if ecu_response.response_code == UDSErrorCodes.responseTooLong:
                logger.error(
                    f"There are too many codes for (sub)mask {mask}. Consider setting --mask with a parameter that excludes one or more of the corresponding bits."
                )
                if split:
                    logger.warning("Trying to fetch the error codes iteratively.")

                    for i in range(8):
                        sub_mask = mask & 2**i

                        if sub_mask > 0:
                            logger.info(f"Trying to fetch with mask {g_repr(sub_mask)}")
                            dtcs.update(await self.fetch_error_codes(sub_mask, False))
            else:
                logger.critical(f"Could not fetch error codes: {ecu_response}; exiting…")
                sys.exit(1)
        else:
            dtcs = ecu_response.dtc_and_status_record

        return dtcs

    async def main(self) -> None:
        dtcs = await self.fetch_error_codes(self.config.mask)

        failed_dtcs: list[list[str]] = []
        uncompleted_dtcs: list[list[str]] = []

        for dtc, error_state in dtcs.items():
            raw_output = f"{dtc:06X} {error_state:02X}"
            bit_string = bin(error_state + 0x100)[
                3:
            ]  # Transform error_state into a bit-string with leading zeros
            table_output = [f"{dtc:06X}", f"{error_state:02X}"] + [
                "X" if b == "1" else "" for b in bit_string
            ]

            # if any kind of test failure
            if error_state & 0xAF:
                logger.warning(raw_output)
                failed_dtcs.append(table_output)
            # if not failed but also not completed yet (i.e. not yet in this cycle or since last clear)
            elif error_state & 0x50:
                logger.result(raw_output)
                uncompleted_dtcs.append(table_output)

        if self.config.show_legend:
            logger.result("")
            self.show_bit_legend()

        if self.config.show_failed:
            logger.result("")
            logger.result("Failed codes:")
            self.show_summary(failed_dtcs)

        if self.config.show_uncompleted:
            logger.result("")
            logger.result("Uncompleted codes:")
            self.show_summary(uncompleted_dtcs)

    def show_bit_legend(self) -> None:
        bit_descriptions = [
            "0 = testFailed: most recent test failed",
            "1 = testFailedThisOperationCycle: failed in current cycle",
            "2 = pendingDTC: failed in current or last cycle",
            "3 = confirmedDTC: failed often enough to be stored in long term memory",
            "4 = testNotCompletedSinceLastClear: not completed since last clear",
            "5 = testFailedSinceLastClear: failed since last clear",
            "6 = testNotCompletedThisOperationCycle: not completed in current cycle",
            "7 = warningIndicatorRequested: existing warning indicators (e.g. lamp, display)",
        ]

        for line in tabulate(
            [[d] for d in bit_descriptions], headers=["bit descriptions"]
        ).splitlines():
            logger.result(line)

    def show_summary(self, dtcs: list[list[str]]) -> None:
        dtcs.sort()

        header = ["DTC", "error state", "0", "1", "2", "3", "4", "5", "6", "7"]

        for line in tabulate(dtcs, headers=header, tablefmt="fancy_grid").splitlines():
            logger.result(line)


class ClearDTCPrimitive(UDSScanner):
    CONFIG_TYPE = ClearDTCPrimitiveConfig
    SHORT_HELP = "Clear the DTCs using the ClearDiagnosticInformation service"

    def __init__(self, config: ClearDTCPrimitiveConfig):
        super().__init__(config)
        self.config: ClearDTCPrimitiveConfig = config

    async def main(self) -> None:
        group_of_dtc: int = self.config.group_of_dtc

        min_group_of_dtc = 0
        max_group_of_dtc = 0xFFFFFF

        if not min_group_of_dtc <= group_of_dtc <= max_group_of_dtc:
            logger.error(
                f"The parameter group_of_dtc must be in the range {g_repr(min_group_of_dtc)}-{g_repr(max_group_of_dtc)}"
            )

        resp = await self.ecu.clear_diagnostic_information(group_of_dtc)

        if isinstance(resp, NegativeResponse):
            logger.error(resp)
        else:
            logger.result("Success")


class ControlDTCPrimitive(UDSScanner):
    CONFIG_TYPE = ControlDTCPrimitiveConfig
    SHORT_HELP = "Stop or resume the setting of DTCs using the ControlDTCSetting service"

    def __init__(self, config: ControlDTCPrimitiveConfig):
        super().__init__(config)
        self.config: ControlDTCPrimitiveConfig = config

    async def main(self) -> None:
        assert isinstance(self.config, ControlDTCPrimitiveConfig)

        if self.config.stop:
            await self.ecu.control_dtc_setting(CDTCSSubFuncs.OFF)
        else:
            await self.ecu.control_dtc_setting(CDTCSSubFuncs.ON)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import contextvars
import importlib.util
import re
import sys
from collections.abc import Awaitable, Callable
from functools import wraps
from types import ModuleType
from typing import Any, ParamSpec, TypeVar

from gallia.log import Loglevel, get_logger

logger = get_logger(__name__)


def auto_int(arg: str) -> int:
    return int(arg, 0)


def strtobool(val: str) -> bool:
    val = val.lower()
    match val:
        case "y" | "yes" | "t" | "true" | "on" | "1":
            return True
        case "n" | "no" | "f" | "false" | "off" | "0":
            return False
        case _:
            raise ValueError(f"invalid truth value {val!r}")


def camel_to_snake(s: str) -> str:
    """Convert a CamelCase string to a snake_case string."""
    # https://stackoverflow.com/a/1176023
    s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s).lower()


def camel_to_dash(s: str) -> str:
    """Convert a CamelCase string to a dash-case string."""
    return camel_to_snake(s).replace("_", "-")


def isotp_addr_repr(a: int) -> str:
    """
    Default string representation of a CAN id.
    """
    return f"{a:02x}"


def can_id_repr(i: int) -> str:
    """
    Default string representation of a CAN id.
    """
    return f"{i:03x}"


def unravel(listing: str) -> list[int]:
    """
    Parses a string representing a one-dimensional list of ranges into an equivalent python data structure.

    Ranges are delimited by hyphens ('-').
    Enumerations are delimited by commas (',').

    Ranges are allowed to overlap and are merged.
    Ranges are always unraveled, which could lead to high memory consumption for distant limits.

    Example: 0,10,8-11
    This would result in [0,8,9,10,11].

    :param listing: The string representation of the one-dimensional list of ranges.
    :return: A list of numbers.
    """

    listing_delimiter = ","
    range_delimiter = "-"
    result = set()

    if listing == "" or listing.isspace():
        return []

    for range_element in listing.split(listing_delimiter):
        if range_delimiter in range_element:
            first_tmp, last_tmp = range_element.split(range_delimiter)
            first = auto_int(first_tmp)
            last = auto_int(last_tmp)

            for element in range(first, last + 1):
                result.add(element)
        else:
            element = auto_int(range_element)
            result.add(element)

    return sorted(result)


def unravel_2d(listing: str) -> dict[int, list[int] | None]:
    """
    Parses a string representing a two-dimensional list of ranges into an equivalent python data structure.

    The outer dimension entries are separated by spaces (' ').
    Inner dimension ranges and outer dimension ranges are separated by colons (':').
    Ranges in both dimensions are delimited by hyphens ('-').
    Enumerations in both dimensions are delimited by commas (',').

    Ranges are allowed to overlap and are merged.
    Ranges are always unraveled, which could lead to high memory consumption for distant limits.
    If a range with only outer dimensions is given, this will result in None for the inner list and overrides other values.

    Example: "1:1,2  1-3:0,2-4  3"
    This would result in {1: [0,1,2,3,4], 2: [0,2,3,4], 3: None}.

    :param listing: The string representation of the two-dimensional list of ranges.
    :return: A mapping of numbers in the outer dimension to numbers in the inner dimension.
    """

    listing_delimiter = " "
    level_delimiter = ":"

    unsorted_result: dict[int, set[int] | None] = {}

    for range_element in listing.split(listing_delimiter):
        if level_delimiter in range_element:
            first_tmp, second_tmp = range_element.split(level_delimiter)
            first = unravel(first_tmp)
            second = unravel(second_tmp)

            for x in first:
                if x not in unsorted_result:
                    unsorted_result[x] = set()

                if (ur := unsorted_result[x]) is not None:
                    for y in second:
                        ur.add(y)
        else:
            first = unravel(range_element)

            for x in first:
                unsorted_result[x] = None

    return {
        x: None if (ur := unsorted_result[x]) is None else sorted(ur)
        for x in sorted(unsorted_result)
    }


T = TypeVar("T")


async def catch_and_log_exception(
    func: Callable[..., Awaitable[T]],
    *args: Any,
    **kwargs: Any,
) -> T | None:
    """Runs an async function. If an exception is raised,
    it will be logged via logger.

    :param logger: an instance of gallia.penlog.Logger
    :param func: a async function object which will be awaited
    :return: None
    """
    try:
        return await func(*args, **kwargs)
    except Exception as e:
        logger.error(f"func {func.__name__} failed: {repr(e)}")
        return None


def lazy_import(name: str) -> ModuleType:
    if name in sys.modules:
        return sys.modules[name]
    spec = importlib.util.find_spec(name)
    if spec is None or spec.loader is None:
        raise ImportError

    loader = importlib.util.LazyLoader(spec.loader)
    spec.loader = loader
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    loader.exec_module(module)
    return module


# TODO: (Re)move these functions
def dump_args(args: Any) -> dict[str, str | int | float]:
    settings = {}
    for key, value in args.__dict__.items():
        match value:
            case str() | int() | float():
                settings[key] = value

    return settings


def get_log_level(cli_level: int) -> Loglevel:
    level = Loglevel.INFO
    if cli_level == 1:
        level = Loglevel.DEBUG
    elif cli_level == 2:
        level = Loglevel.TRACE
    return level


def get_file_log_level(args: Any) -> Loglevel:
    level = Loglevel.DEBUG
    if hasattr(args, "trace_log"):
        if args.trace_log:
            level = Loglevel.TRACE
    elif hasattr(args, "verbose"):
        if args.verbose >= 2:
            level = Loglevel.TRACE
    return level


CONTEXT_SHARED_VARIABLE = "logger_name"
context: contextvars.ContextVar[tuple[str, str | None]] = contextvars.ContextVar(
    CONTEXT_SHARED_VARIABLE
)


def set_task_handler_ctx_variable(
    logger_name: str,
    task_name: str | None = None,
) -> contextvars.Context:
    ctx = contextvars.copy_context()
    ctx.run(context.set, (logger_name, task_name))
    return ctx


def handle_task_error(fut: asyncio.Future[Any]) -> None:
    logger_name, task_name = context.get((__name__, "Task"))
    logger = get_logger(logger_name)
    if logger.name is __name__:
        logger.warning(f"BUG: {fut} had no context '{CONTEXT_SHARED_VARIABLE}' set")
        logger.warning("BUG: please fix this to ensure the task name is logged correctly")

    try:
        fut.result()
    except BaseException as e:
        task_name = task_name if task_name is not None else "Task"

        # Debug level is enough, since our aim is only to consume the stack trace
        logger.debug(f"{task_name} ended with error: {e!r}")


P = ParamSpec("P")


def supports_platform(*platform: str) -> Callable[[Callable[P, T]], Callable[P, T]]:
    def decorator(function: Callable[P, T]) -> Callable[P, T]:
        @wraps(function)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            supported = False
            for p in platform:
                if sys.platform == p:
                    supported = True
                    break
            if supported is False:
                raise NotImplementedError(
                    f'`{function.__name__}()` is not supported on: "{sys.platform}"'
                )

            return function(*args, **kwargs)

        return wrapper

    return decorator
